/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.Collections.Generic;

namespace plexdata.LineCounter
{
    // TODO: Add more properties if needed.

    public class Settings : ICloneable
    {
        private static Settings instance = null;

        public static Settings Instance
        {
            get
            {
                if (Settings.instance == null)
                {
                    Settings.instance = new Settings();
                }
                return Settings.instance;
            }
        }

        private Settings()
            : base()
        {
            this.ResetDefaults();
        }

        private Settings(Settings clone)
            : this()
        {
            this.FromClone(clone);
        }

        public bool Save()
        {
            return plexdata.Generic.Serializer.Save<Settings>(
                Settings.ConfigurationFilename,
                Settings.instance);
        }

        public bool Load()
        {
            return plexdata.Generic.Serializer.Load<Settings>(
                Settings.ConfigurationFilename,
                out Settings.instance);
        }

        public General General { get; set; }

        public MainWindow MainWindow { get; set; }

        public FileSelectorWindow FileSelectorWindow { get; set; }

        public FolderSelectorWindow FolderSelectorWindow { get; set; }

        public FileFilterGroupWindow FileFilterGroupWindow { get; set; }

        public void ResetDefaults()
        {
            this.General = new General();
            this.MainWindow = new MainWindow();
            this.FileSelectorWindow = new FileSelectorWindow();
            this.FolderSelectorWindow = new FolderSelectorWindow();
            this.FileFilterGroupWindow = new FileFilterGroupWindow();
        }

        public void FromClone(object clone)
        {
            if (clone == null)
            {
                throw new ArgumentNullException("clone");
            }

            if (clone is Settings)
            {
                this.General = new General((clone as Settings).General);
                this.MainWindow = new MainWindow((clone as Settings).MainWindow);
                this.FileSelectorWindow = new FileSelectorWindow((clone as Settings).FileSelectorWindow);
                this.FolderSelectorWindow = new FolderSelectorWindow((clone as Settings).FolderSelectorWindow);
                this.FileFilterGroupWindow = new FileFilterGroupWindow((clone as Settings).FileFilterGroupWindow);
            }
            else
            {
                throw new InvalidCastException();
            }
        }

        #region ICloneable member implementation

        public object Clone()
        {
            return new Settings(this);
        }

        #endregion // ICloneable member implementation

        private static string ConfigurationFilename
        {
            get
            {
                string path = Path.GetFullPath(Assembly.GetExecutingAssembly().Location);
                return Path.GetDirectoryName(path) + Path.DirectorySeparatorChar +
                    Path.GetFileNameWithoutExtension(path) + ".xml";
            }
        }
    }

    public abstract class WindowSettings
    {
        private FormWindowState state;
        private Point location;
        private Size size;

        public WindowSettings()
            : base()
        {
            this.State = FormWindowState.Normal;
        }

        public FormWindowState State
        {
            get { return this.state; }
            set { this.state = value; }
        }

        public Point Location
        {
            get
            {
                if (this.location == null)
                {
                    return this.DefaultLocation;
                }
                else
                {
                    return this.location;
                }
            }
            set
            {
                if (value == null)
                {
                    this.location = this.DefaultLocation;
                }
                else
                {
                    this.location = this.EnsureScreenLocation(value);
                }
            }
        }

        public Size Size
        {
            get
            {
                if (this.size == null)
                {
                    return this.DefaultSize;
                }
                else
                {
                    return this.size;
                }
            }
            set
            {
                if (value == null)
                {
                    this.size = this.DefaultSize;
                }
                else
                {
                    this.size = new Size(value.Width, value.Height);
                }
            }
        }

        [XmlIgnore]
        public abstract FormWindowState DefaultState { get; }

        [XmlIgnore]
        public abstract Point DefaultLocation { get; }

        [XmlIgnore]
        public abstract Size DefaultSize { get; }

        public virtual void ResetDefaults()
        {
            this.State = this.DefaultState;
            this.Location = this.DefaultLocation;
            this.Size = this.DefaultSize;
        }

        public virtual void FromClone(object clone)
        {
            if (clone == null)
            {
                throw new ArgumentNullException("clone");
            }

            if (clone is WindowSettings)
            {
                this.State = (clone as WindowSettings).State;

                this.Location = new Point(
                    (clone as WindowSettings).Location.X,
                    (clone as WindowSettings).Location.Y);

                this.Size = new Size(
                    (clone as WindowSettings).Size.Width,
                    (clone as WindowSettings).Size.Height);
            }
            else
            {
                throw new InvalidCastException();
            }
        }

        private Point EnsureScreenLocation(Point location)
        {
            if (this.State == FormWindowState.Maximized || this.State == FormWindowState.Minimized)
            {
                // In this case let the form handle resizing.
                return location;
            }

            foreach (Screen screen in Screen.AllScreens)
            {
                if (screen.Bounds.Contains(new Rectangle(location, this.Size)))
                {
                    return location; // Location is still within bounds.
                }
            }

            Rectangle bounds = Screen.PrimaryScreen.WorkingArea;

            // Adjust X location to be on the primary screen!

            int x = 0;
            if (location.X < bounds.Left)
            {
                x = bounds.Left;
            }
            else if (location.X + this.Size.Width > bounds.Left + bounds.Right)
            {
                x = bounds.Right - this.Size.Width;
            }
            else
            {
                x = location.X;
            }

            // adjust Y location to be on the primary screen!

            int y = 0;
            if (location.Y < bounds.Top)
            {
                y = bounds.Top;
            }
            else if (location.Y + this.Size.Height > bounds.Top + bounds.Bottom)
            {
                y = bounds.Bottom - this.Size.Height;
            }
            else
            {
                y = location.Y;
            }

            return new Point(x, y);
        }
    }

    public class MainWindow : WindowSettings, ICloneable
    {
        public MainWindow()
            : base()
        {
            this.ResetDefaults();
        }

        public MainWindow(MainWindow clone)
            : this()
        {
            this.FromClone(clone);
        }

        #region Overwritten abstract member functions.

        public override FormWindowState DefaultState
        {
            get { return FormWindowState.Normal; }
        }

        public override Point DefaultLocation
        {
            get { return new Point(200, 200); }
        }

        public override Size DefaultSize
        {
            get { return new Size(640, 480); }
        }

        #endregion // Overwritten abstract member functions.

        #region Overwritten inherited member functions.

        public override void ResetDefaults()
        {
            base.ResetDefaults();
        }

        public override void FromClone(object clone)
        {
            base.FromClone(clone);
        }

        #endregion // Overwritten inherited member functions.

        #region ICloneable member implementation

        public object Clone()
        {
            return new MainWindow(this);
        }

        #endregion // ICloneable member implementation
    }

    public class FileSelectorWindow : WindowSettings, ICloneable
    {
        public FileSelectorWindow()
            : base()
        {
            this.ResetDefaults();
        }

        public FileSelectorWindow(FileSelectorWindow clone)
            : this()
        {
            this.FromClone(clone);
        }

        #region Overwritten abstract member functions.

        public override FormWindowState DefaultState
        {
            get { return FormWindowState.Normal; }
        }

        public override Point DefaultLocation
        {
            get { return new Point(200, 200); }
        }

        public override Size DefaultSize
        {
            get { return new Size(600, 400); }
        }

        #endregion // Overwritten abstract member functions.

        #region Overwritten inherited member functions.

        public override void ResetDefaults()
        {
            base.ResetDefaults();
        }

        public override void FromClone(object clone)
        {
            base.FromClone(clone);
        }

        #endregion // Overwritten inherited member functions.

        #region ICloneable member implementation

        public object Clone()
        {
            return new FileSelectorWindow(this);
        }

        #endregion // ICloneable member implementation
    }

    public class FolderSelectorWindow : WindowSettings, ICloneable
    {
        public FolderSelectorWindow()
            : base()
        {
            this.ResetDefaults();
        }

        public FolderSelectorWindow(FolderSelectorWindow clone)
            : this()
        {
            this.FromClone(clone);
        }

        #region Overwritten abstract member functions.

        public override FormWindowState DefaultState
        {
            get { return FormWindowState.Normal; }
        }

        public override Point DefaultLocation
        {
            get { return new Point(200, 200); }
        }

        public override Size DefaultSize
        {
            get { return new Size(600, 400); }
        }

        #endregion // Overwritten abstract member functions.

        #region Overwritten inherited member functions.

        public override void ResetDefaults()
        {
            base.ResetDefaults();
        }

        public override void FromClone(object clone)
        {
            base.FromClone(clone);
        }

        #endregion // Overwritten inherited member functions.

        #region ICloneable member implementation

        public object Clone()
        {
            return new FolderSelectorWindow(this);
        }

        #endregion // ICloneable member implementation
    }

    public class FileFilterGroupWindow : WindowSettings, ICloneable
    {
        public FileFilterGroupWindow()
            : base()
        {
            this.ResetDefaults();
        }

        public FileFilterGroupWindow(FileFilterGroupWindow clone)
            : this()
        {
            this.FromClone(clone);
        }

        #region Overwritten abstract member functions.

        public override FormWindowState DefaultState
        {
            get { return FormWindowState.Normal; }
        }

        public override Point DefaultLocation
        {
            get { return new Point(200, 200); }
        }

        public override Size DefaultSize
        {
            get { return new Size(550, 403); }
        }

        #endregion // Overwritten abstract member functions.

        #region Overwritten inherited member functions.

        public override void ResetDefaults()
        {
            base.ResetDefaults();
        }

        public override void FromClone(object clone)
        {
            base.FromClone(clone);
        }

        #endregion // Overwritten inherited member functions.

        #region ICloneable member implementation

        public object Clone()
        {
            return new FileFilterGroupWindow(this);
        }

        #endregion // ICloneable member implementation
    }

    public class General : ICloneable
    {
        public General()
            : base()
        {
            this.ResetDefaults();
        }

        public General(General clone)
            : this()
        {
            this.FromClone(clone);
        }

        private string lastEnvironment = String.Empty;
        public string LastEnvironment
        {
            get { return this.lastEnvironment; }
            set { this.lastEnvironment = (value == null) ? String.Empty : value; }
        }

        private string lastReportFile = String.Empty;
        public string LastReportFile
        {
            get { return this.lastReportFile; }
            set { this.lastReportFile = (value == null) ? String.Empty : value; }
        }

        public void ResetDefaults()
        {
            this.LastEnvironment = String.Empty;
            this.LastReportFile = String.Empty;
        }

        public void FromClone(object clone)
        {
            if (clone == null)
            {
                throw new ArgumentNullException("clone");
            }

            if (clone is General)
            {
                this.LastEnvironment = (clone as General).LastEnvironment.Clone() as string;
                this.LastReportFile = (clone as General).LastReportFile.Clone() as string;
            }
            else
            {
                throw new InvalidCastException();
            }
        }

        #region ICloneable member implementation

        public object Clone()
        {
            return new General(this);
        }

        #endregion // ICloneable member implementation
    }
}
