﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.Text;
using System.Xml.Serialization;
using System.Collections.Generic;

namespace plexdata.LineCounter
{
    public class CommentScheme : ICloneable, IEquatable<CommentScheme>, IComparable<CommentScheme>
    {
        private const string EOL = "EOL";
        private const char SPC = ' ';
        private const char TAB = '\t';

        public CommentScheme()
            : this(null, null, false)
        {
        }

        public CommentScheme(string open)
            : this(open, null, false)
        {
        }

        public CommentScheme(string open, bool whitespace)
            : this(open, null, whitespace)
        {
        }

        public CommentScheme(string open, string close)
            : this(open, close, false)
        {
        }

        public CommentScheme(string open, string close, bool whitespace)
            : base()
        {
            this.Open = open;
            this.Close = close;
            this.Whitespace = whitespace;
        }

        private string open = String.Empty;
        public string Open
        {
            get
            {
                return this.open;
            }
            set
            {
                if (value == null)
                {
                    this.open = String.Empty;
                }
                else
                {
                    this.open = value.Trim();
                }
                this.IgnoreCase = this.CanIgnoreCase();
            }
        }

        private string close = String.Empty;
        public string Close
        {
            get
            {
                return this.close;
            }
            set
            {
                if (value == null)
                {
                    this.close = String.Empty;
                }
                // Adjust close tag if EOL was given!
                else if (String.Compare(value, EOL, true) == 0)
                {
                    this.close = String.Empty;
                }
                else
                {
                    this.close = value.Trim();
                }

                this.IgnoreCase = this.CanIgnoreCase();
            }
        }

        public bool Whitespace { get; set; }

        [XmlIgnore]
        public bool IsOpen
        {
            get { return this.open.Length > 0; }
        }

        [XmlIgnore]
        public bool IsClose
        {
            get { return this.close.Length > 0; }
        }

        [XmlIgnore]
        public bool IgnoreCase { get; private set; }

        [XmlIgnore]
        public bool EndOfLine { get { return !this.IsClose; } }

        public bool IsMatching(StringBuilder buffer, char linefeed, int maximum, ref int index, out int length)
        {
            length = 0;

            bool matching = false;
            int posOpen = this.GetOpenPosition(buffer, index, linefeed, maximum);
            if (posOpen >= 0)
            {
                // Points onto end of open tag! Therefore, adjust offset.
                int offset = posOpen - this.open.Length + 1;
                if (this.Whitespace) { offset -= 1; }

                // Found matching is not at expected location!
                if (offset != index) { return false; }

                // Start searching for the close tag at the beginning 
                // of the open tag to prevent problems in finding the 
                // close tag in case of a mismatching comment scheme.
                int posClose = this.GetClosePosition(buffer, offset, linefeed);
                if (posClose < 0)
                {
                    throw new IndexOutOfRangeException(String.Format(
                        "Comment open [{0}] has no corresponding comment close [{1}]!",
                        this.Open, this.IsClose ? this.Close : EOL));
                }
                else if (posClose > posOpen)
                {
                    matching = true;
                    index = offset;

                    // Points onto start of close tag! Therefore, adjust 
                    // length. Note in case of the comment ends at the 
                    // line end (index + length) points onto the linefeed. 
                    // Otherwise (index + length) points onto first character 
                    // behinde closing tag.
                    length = posClose + this.close.Length - offset;
                }
            }
            return matching;
        }

        public object Clone()
        {
            return new CommentScheme(this.Open, this.Close, this.Whitespace);
        }

        public int CompareTo(CommentScheme other)
        {
            if (other != null)
            {
                if (this.open.Length > other.open.Length)
                {
                    // If length of "this" is greater than length 
                    // of "other" then "this" is greater.
                    return 1;
                }
                else if (this.open.Length < other.open.Length)
                {
                    // If length of "this" is less than length of 
                    // "other" then "this" is less.
                    return -1;
                }
                else
                {
                    // Otherwise length is equal and checking 
                    // the content is required.
                    for (int idx = 0; idx < this.open.Length; idx++)
                    {
                        char left = this.open[idx];
                        char right = other.open[idx];

                        if (this.IgnoreCase)
                        {
                            left = Char.ToLower(left);
                            right = Char.ToLower(right);
                        }

                        if (left > right)
                        {
                            return 1;
                        }
                        else if (left < right)
                        {
                            return -1;
                        }
                    }

                    // Up to here the open tag is equal. 
                    // Now check close tag in the same way.

                    if (this.close.Length > other.close.Length)
                    {
                        return 1;
                    }
                    else if (this.close.Length < other.close.Length)
                    {
                        return -1;
                    }
                    else
                    {
                        for (int idx = 0; idx < this.close.Length; idx++)
                        {
                            char left = this.close[idx];
                            char right = other.close[idx];

                            if (this.IgnoreCase)
                            {
                                left = Char.ToLower(left);
                                right = Char.ToLower(right);
                            }

                            if (left > right)
                            {
                                return 1;
                            }
                            else if (left < right)
                            {
                                return -1;
                            }
                        }
                    }

                    // Up to here the close tag is equal too. 

                    // Now check value of "whitespace", 
                    // "true" is greater than "false".
                    if (this.Whitespace && !other.Whitespace)
                    {
                        return 1;
                    }
                    else if (!this.Whitespace && other.Whitespace)
                    {
                        return -1;
                    }
                    else
                    {
                        return 0;
                    }
                }
            }

            // If "other" is null then "this" is definitely greater.
            return 1;
        }

        public bool Equals(CommentScheme other)
        {
            if (other != null)
            {
                if (this.open.Length == other.open.Length)
                {
                    for (int idx = 0; idx < this.open.Length; idx++)
                    {
                        char left = this.open[idx];
                        char right = other.open[idx];

                        if (this.IgnoreCase)
                        {
                            left = Char.ToLower(left);
                            right = Char.ToLower(right);
                        }

                        if (left != right) { return false; }
                    }

                    if (this.close.Length == other.close.Length)
                    {
                        for (int idx = 0; idx < this.close.Length; idx++)
                        {
                            char left = this.close[idx];
                            char right = other.close[idx];

                            if (this.IgnoreCase)
                            {
                                left = Char.ToLower(left);
                                right = Char.ToLower(right);
                            }

                            if (left != right) { return false; }
                        }
                    }
                    else
                    {
                        return false;
                    }

                    // Up to here "this" and "other" are equal. 
                    // Therefore, check white spaces.
                    return (this.Whitespace == other.Whitespace);
                }
                else
                {
                    return false;
                }
            }
            return false;
        }

        public string ToString(bool tagsOnly)
        {
            string result = String.Empty;
            result += this.IsOpen ? this.open : EOL;
            result = result.PadRight(6, SPC);
            result += this.IsClose ? this.close : EOL;
            result = result.PadRight(12, SPC);
            if (!tagsOnly) { result += this.Whitespace.ToString().ToLower(); }
            return result;
        }

        public override string ToString()
        {
            return this.ToString(false);
        }

        private bool CanIgnoreCase()
        {
            // Ignore case is needed as soon as one character is a letter!

            for (int idx = 0; idx < this.open.Length; idx++)
            {
                if (Char.IsLetter(this.open[idx])) { return true; }
            }

            for (int idx = 0; idx < this.close.Length; idx++)
            {
                if (Char.IsLetter(this.close[idx])) { return true; }
            }
            return false;
        }

        private int GetOpenPosition(StringBuilder buffer, int offset, char linefeed, int maximum)
        {
            int result = -1;
            int start = offset;

            if (this.IsOpen)
            {
                for (; offset < buffer.Length; offset++)
                {
                    // Stop searching if maximum has been reached.
                    if (offset >= start + maximum) { return -1; }

                    char left = buffer[offset];
                    char right = this.open[0];

                    if (this.IgnoreCase)
                    {
                        left = Char.ToLower(left);
                        right = Char.ToLower(right);
                    }

                    // Check other characters only if first character matches.
                    if (left == right)
                    {
                        // This character has been checked already!
                        offset++;

                        // Try to validate open tag!
                        bool validated = true;
                        for (int inner = 1; validated && inner < this.open.Length && offset < buffer.Length; inner++)
                        {
                            left = buffer[offset];
                            right = this.open[inner];

                            if (this.IgnoreCase)
                            {
                                left = Char.ToLower(left);
                                right = Char.ToLower(right);
                            }

                            validated = validated & (left == right);
                            if (validated) { offset++; }
                        }

                        if (this.Whitespace)
                        {
                            validated = validated & (buffer[offset] == SPC || buffer[offset] == TAB);
                            if (validated) { offset++; }
                        }

                        if (!validated)
                        {
                            continue;
                        }
                        else
                        {
                            // Adjust result to point onto 
                            // last open tag character.
                            result = offset - 1;
                            break;
                        }
                    }
                    else
                    {
                        // Stop searching if current character doesn't fit.
                        return -1;
                    }
                }
            }
            return result;
        }

        private int GetClosePosition(StringBuilder buffer, int offset, char linefeed)
        {
            int result = -1;

            string closing = this.IsClose ? this.close : linefeed.ToString();

            for (; offset < buffer.Length; offset++)
            {
                char left = buffer[offset];
                char right = closing[0];

                if (this.IgnoreCase)
                {
                    left = Char.ToLower(left);
                    right = Char.ToLower(right);
                }

                // Check other characters only if first character matches.
                if (left == right)
                {
                    // Save current position for jumping back.
                    int position = offset;

                    // This character has been checked already!
                    offset++;

                    // Try validating close tag!
                    bool validated = true;
                    for (int inner = 1; validated && inner < closing.Length && offset < buffer.Length; inner++)
                    {
                        left = buffer[offset];
                        right = closing[inner];

                        if (this.IgnoreCase)
                        {
                            left = Char.ToLower(left);
                            right = Char.ToLower(right);
                        }

                        validated = validated & (left == right);
                        if (validated) { offset++; }
                    }

                    if (!validated)
                    {
                        // Go back to last known position.
                        offset = position;
                        continue;
                    }
                    else
                    {
                        // Adjust result to point onto 
                        // first close tag character.
                        result = offset - closing.Length;
                        break;
                    }
                }
            }
            return result;
        }
    }

    public class CommentSchemeCollection : List<CommentScheme>, ICloneable
    {
        public CommentSchemeCollection()
            : base()
        {
        }

        public new void Add(CommentScheme value)
        {
            // Check given value but suppress duplicates...
            if (value != null && !base.Contains(value))
            {
                base.Add(value);
            }
        }

        public new void AddRange(IEnumerable<CommentScheme> collection)
        {
            if (collection != null)
            {
                foreach (CommentScheme item in collection)
                {
                    this.Add(item);
                }
            }
        }

        public void Sort(bool ascending)
        {
            base.Sort();
            if (!ascending)
            {
                base.Reverse();
            }
        }

        public object Clone()
        {
            CommentSchemeCollection clone = new CommentSchemeCollection();
            for (int index = 0; index < this.Count; index++)
            {
                clone.Add(this[index].Clone() as CommentScheme);
            }
            return clone;
        }
    }
}
