﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Globalization;
using System.Collections.Generic;

namespace plexdata.LineCounter
{
    public class CommentFindings
    {
        public CommentFindings(CommentScheme scheme)
            : base()
        {
            if (scheme != null)
            {
                this.Scheme = (CommentScheme)scheme.Clone();
                this.PureCommentCount = 0;
                this.PureCommentLines = 0;
                this.CodeCommentCount = 0;
                this.CodeCommentLines = 0;
            }
            else
            {
                throw new ArgumentNullException();
            }
        }

        public CommentScheme Scheme { get; private set; }

        // Number of pure comment type occurrences (no code in front and no code behind). 
        // Amount: Less or equal to number of pure comment lines.
        public int PureCommentCount { get; set; }

        // Number of lines of pure comment types (no code in front and no code behind). 
        // In case of single line comments equal to number of pure comment types.
        public int PureCommentLines { get; set; }

        // Number of code comment type occurrences (code in front or code behind). 
        // In case of multiple lines inside a comment less than number of code comment lines.
        public int CodeCommentCount { get; set; }

        // Number of lines of code comment types (code in front or code behind). 
        // In case of multiple hits in one line less than the number of code comment types.
        public int CodeCommentLines { get; set; }

        public override string ToString()
        {
            return this.Scheme.ToString(true) +
                " Pure: " + this.PureCommentCount + " (" + this.PureCommentLines + ")" +
                " Code: " + this.CodeCommentCount + " (" + this.CodeCommentLines + ")";
        }
    }

    public enum Linefeed
    {
        // Happens when a file contains neither CRs nor LFs ("single line file").
        Unknown,
        // Happens when a file contains CRs followed by LFs.
        Windows,
        // Happens when a file only contains LFs.
        Unix,
    }

    public class ContentScanner
    {
        // General Remarks to this Comment Count Algorithm
        // 
        // After a long analysis phase followed by various tries it came up that counting source code 
        // comments is not as easy as it seems on the first glance. This is because of the fact that 
        // many different comment types as well as their combinations are possible inside one single 
        // code file. Furthermore, answering the question "What needs to be counted?" was not easy and 
        // became clear during the implementation. Therefore, the search algorithm implemented here 
        // looks a little bit complicated but it works, at least for the combinations of single-line 
        // (ending on EOL) and multi-line comments listed below. This means, don't touch...
        // 
        // If there are only white spaces between two columns then this line is treated as comment line!
        // 
        // +-----------------------------------+---------------+---------------+-------+
        // | Comment Type Combinations         | Pure          | Code          |  LOC  |
        // |                                   +-------+-------+-------+-------+       |
        // |                                   | Count | Lines | Count | Lines |       |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | //...                             |   1   |   1   |       |       |  n/a  |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...*/                           |   1   |   1   |       |       |  n/a  |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | code//...                         |       |       |   1   |   1   |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | code/*...*/                       |       |       |   1   |   1   |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...*/code                       |       |       |   1   |   1   |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | code/*...*/code                   |       |       |   1   |   1   |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...*///...                      |       |       |   2   |   1   |  n/a  | <- Line assigned to "//".
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...*//*...*/                    |       |       |   2   |   1   |  n/a  |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...*/code/*...*/                |       |       |   2   |       |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...*/code/*...*/code/*...*/     |       |       |   3   |       |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...*/code//...                  |       |       |   2   |       |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...                             |   1   |   1   |       |       |  n/a  |
        // |   ...                             |       |   1   |       |       |       |
        // |   ...*/                           |       |   1   |       |       |       |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | code/*...                         |       |       |   1   |       |   1   |
        // |       ...                         |       |       |       |   1   |       |
        // |       ...*/                       |       |       |       |   1   |       |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...                             |       |       |   1   |   1   |       |
        // |   ...                             |       |       |       |   1   |       |
        // |   ...*/code                       |       |       |       |       |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | code/*...                         |       |       |   1   |       |   1   |
        // |       ...                         |       |       |       |   1   |       |
        // |       ...*/code                   |       |       |       |       |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...                             |       |       |   1   |   1   |  n/a  |
        // |   ...                             |       |       |       |   1   |       |
        // |   ...*///...                      |       |       |   1   |   1   |       | <- Line assigned to "//".
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | code/*...                         |       |       |   1   |       |   1   |
        // |       ...                         |       |       |       |   1   |       |
        // |       ...*///...                  |       |       |   1   |   1   |       | <- Line assigned to "//".
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | /*...                             |       |       |   1   |   1   |       |
        // |   ...                             |       |       |       |   1   |       |
        // |   ...*/code//...                  |       |       |   1   |       |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        // | code/*...                         |       |       |   1   |       |   1   |
        // |       ...                         |       |       |       |   1   |       |
        // |       ...*/code//...              |       |       |   1   |       |   1   |
        // +-----------------------------------+-------+-------+-------+-------+-------+
        //

        private const char SPC = ' ';
        private const char TAB = '\t';
        private const char NIL = '\0'; // Replacement for comment content.
        private const char LF = '\n';  // Used in Windows and Unix files.
        private const char CR = '\r';  // Used only in Windows files.

        public ContentScanner(string filename, CommentSchemeCollection schemes)
            : base()
        {
            this.Filename = filename;
            this.Findings = this.BuildFindings(schemes);
            this.ResetProperties();
            this.Exception = null;
        }

        public string Filename { get; private set; }

        public List<CommentFindings> Findings { get; private set; }

        public TimeSpan ScanDuration { get; private set; }

        public int TotalLines { get; private set; }

        public int EmptyLines { get; private set; }

        public int CommentLines { get; private set; }

        public int LinesWithCode { get; private set; }

        public int StringLiterals { get; private set; }

        public int CharacterLiterals { get; private set; }

        public Encoding Encoding { get; private set; }

        public Linefeed Linefeed { get; private set; }

        public Exception Exception { get; private set; }

        public bool Scan(LiteralScheme stringLiteral, LiteralScheme characterLiteral)
        {
            bool success = false;
            try
            {
                // Reset all properties as first.
                this.ResetProperties();
                this.Exception = null;

                int position = 0;           // Current position within content.
                int stringLiterals = 0;     // Total number of string literals.
                int characterLiterals = 0;  // Total number of character literals.

                // Save current start time (loading of content included).
                long startTime = DateTime.Now.Ticks;

                StringBuilder builder = this.LoadContent();
                while (position < builder.Length)
                {
                    if (this.IsLiteralOpen(builder, position, stringLiteral))
                    {
                        // Move to the end of this string literal.
                        position = this.MoveLiteralEnd(builder, position, stringLiteral, ref stringLiterals);
                    }
                    else if (this.IsLiteralOpen(builder, position, characterLiteral))
                    {
                        // Move to the end of this character literal.
                        position = this.MoveLiteralEnd(builder, position, characterLiteral, ref characterLiterals);
                    }
                    else
                    {
                        // Try processing all matching comment schemes.
                        position = this.ProcessComments(builder, position);

                        // It is necessary to check for empty lines after processing 
                        // comments because an empty line could occur inside a comment 
                        // and in this case this empty line belongs to the comment.
                        if (this.IsLineEmpty(builder, position))
                        {
                            this.EmptyLines++;
                        }
                    }
                }

                // Now update all yet missing properties.
                foreach (CommentFindings current in this.Findings)
                {
                    this.CommentLines +=
                        current.PureCommentLines + current.CodeCommentLines;
                }

                this.LinesWithCode = this.TotalLines - (this.EmptyLines + this.CommentLines);
                this.StringLiterals = stringLiterals;
                this.CharacterLiterals = characterLiterals;

                // Last but not least, save current duration.
                this.ScanDuration = new TimeSpan(DateTime.Now.Ticks - startTime);

                success = true;
            }
            catch (Exception exception)
            {
                this.ResetProperties();
                this.Exception = exception;
                Debug.WriteLine(exception);
            }
            return success;
        }

        private int MaxLength { get; set; }

        private void ResetProperties()
        {
            this.ScanDuration = new TimeSpan();
            this.TotalLines = 0;
            this.EmptyLines = 0;
            this.CommentLines = 0;
            this.LinesWithCode = 0;
            this.StringLiterals = 0;
            this.CharacterLiterals = 0;
            this.Encoding = null;
            this.Linefeed = Linefeed.Unknown;
            foreach (CommentFindings current in this.Findings)
            {
                current.PureCommentCount = 0;
                current.PureCommentLines = 0;
                current.CodeCommentCount = 0;
                current.CodeCommentLines = 0;
            }
        }

        private List<CommentFindings> BuildFindings(CommentSchemeCollection schemes)
        {
            List<CommentFindings> result = new List<CommentFindings>();
            if (schemes != null)
            {
                CommentSchemeCollection clone = schemes.Clone() as CommentSchemeCollection;
                // Sort list descending to ensure that for instance
                // comment tag "/**" appears is in front of "/*".
                clone.Sort(false);

                foreach (CommentScheme item in clone)
                {
                    // Determine the maximum length of the Open tag.
                    if (this.MaxLength < item.Open.Length)
                    {
                        this.MaxLength = item.Open.Length;
                    }

                    result.Add(new CommentFindings(item));
                }
            }
            return result;
        }

        private StringBuilder LoadContent()
        {
            StringBuilder builder = null;

            int totalLines = 0;

            using (FileStream stream = new FileStream(this.Filename, FileMode.Open, FileAccess.Read, FileShare.Read))
            using (TextReader reader = new StreamReader(stream, true))
            {
                if (stream.Length > int.MaxValue)
                {
                    throw new NotSupportedException(String.Format(
                        "Files with a size greater than {0} MiB are not supported.",
                        (int.MaxValue / (1024 * 1024)).ToString("N0", NumberFormatInfo.CurrentInfo)));
                }

                string linefeed = LF.ToString();

                builder = new StringBuilder((int)stream.Length);

                while (reader.Peek() != -1)
                {
                    // If the last line contains only spaces then this line is an empty line.
                    builder.Append(reader.ReadLine().Trim() + linefeed);
                    totalLines++;
                }

                // Save number of total file lines.
                this.TotalLines = totalLines;

                // Save current file encoding.
                this.Encoding = (reader as StreamReader).CurrentEncoding;

                // Try to determine current Line Break type.
                this.Linefeed = Linefeed.Unknown;

                if (stream.CanSeek)
                {
                    stream.Position = 0;
                    byte[] array = new byte[this.Encoding.GetByteCount(new char[] { SPC })];

                    while (stream.Position < stream.Length)
                    {
                        stream.Read(array, 0, array.Length);

                        char[] helper = this.Encoding.GetChars(array);
                        if (helper.Length > 0)
                        {
                            if (helper[0] == CR)
                            {
                                // CR found first, it must be a Windows file.
                                this.Linefeed = Linefeed.Windows;
                                break; // Done!
                            }
                            else if (helper[0] == LF)
                            {
                                // No CR found beforehand, it must be a Unix file.
                                this.Linefeed = Linefeed.Unix;
                                break; // Done!
                            }
                        }
                    }
                }
            }

            return builder;
        }

        private bool IsLineEmpty(StringBuilder builder, int position)
        {
            // Empty line is true if this condition fits because lines are trimmed already!
            // position==0 AND builder[position]==LF OR builder[position-1]==LF AND builder[position]==LF; 
            return (position < builder.Length && builder[position] == LF &&
                   (position == 0 || builder[position - 1] == LF));
        }

        private bool IsLiteralOpen(StringBuilder builder, int position, LiteralScheme literal)
        {
            bool matches = true;

            // Ignore empty literals!
            if (literal.Marker.Length == 0) { return false; }

            // A literal is opened when current buffer content exactly fits the literal marker.
            string marker = literal.Marker;
            for (int index = 0; matches && index < marker.Length; index++)
            {
                // Cannot be evaluated (not yet tested)!
                if (position + index > builder.Length) { return false; }

                matches = matches & (marker[index] == builder[position + index]);
            }
            return matches;
        }

        private int MoveLiteralEnd(StringBuilder builder, int position, LiteralScheme literal, ref int count)
        {
            bool matching = false;
            bool escaping = false;

            // Prepare marker and escape tags.
            string marker = literal.Marker;
            string escape = literal.Escape;

            // Save given buffer offset to be able 
            // to go back in case of an error.
            int offset = position;

            // Skip complete marker tag.
            position += marker.Length;

            while (position < builder.Length && builder[position] != LF)
            {
                if (builder[position] == escape[0])
                {
                    // Assume escape mode.
                    escaping = true;

                    // Save comparison start position and move to following character 
                    // (because current character is evaluated already).
                    int start = position++;

                    // Loop until each escape character has been skipped.
                    while (escaping && position < builder.Length && position - start < escape.Length)
                    {
                        // Check if current character fits too.
                        escaping = escaping & (builder[position] == escape[position - start]);

                        // Move to next character but only if current character fits.
                        if (escaping) { position++; }
                    }
                }
                // Check if current character equals first close tag character.
                else if (builder[position] == marker[0])
                {
                    // Assume comment close tag matching.
                    matching = true;

                    // Save comparison start position and move to following character 
                    // (because current character is evaluated already).
                    int start = position++;

                    // Loop until each marker tag character has been skipped.
                    while (matching && position < builder.Length && position - start < marker.Length)
                    {
                        // Check if current character fits too.
                        matching = matching & (builder[position] == marker[position - start]);

                        // Move to next character but only if current character fits.
                        if (matching) { position++; }
                    }

                    // End outer loop if closing tag has been found (but only if not in escape mode).
                    if (matching && !escaping) { break; }

                    // Reset dependent values when in escape mode.
                    if (escaping) { escaping = false; matching = false; }
                }
                else
                {
                    position++;
                }
            }

            if (matching)
            {
                count++; // Increment result value if matching.
            }
            else
            {
                // Go back to last known position, 
                // but auto-move to next character.
                position = offset + 1;
            }
            return position;
        }

        private int ProcessComments(StringBuilder builder, int position)
        {
            int length = 0;      // Length of the found comment type.
            int lines = 0;       // Number of lines inside this comment!
            bool before = false; // Something is available in front of this comment.
            bool behind = false; // Something is available behind this comment.

            foreach (CommentFindings current in this.Findings)
            {
                if (current.Scheme.IsMatching(builder, LF, this.MaxLength, ref position, out length))
                {
                    // Determine number of lines within this comment.
                    for (int index = position; index < position + length; index++)
                    {
                        // REMARK:  Because of an unmanageable complexity it is unfortunately 
                        //          impossible to determine the real number of existing code 
                        //          lines in a reliable way. Therefore, replace all unimportant 
                        //          information with a placeholder, but keep the information 
                        //          about line feeds. See method Scan() for more information.
                        if (builder[index] == LF)
                        {
                            lines++;
                        }
                        else
                        {
                            // And by the way, replace all comment characters to be able 
                            // to determine the number of code lines afterwards.
                            builder.Replace(builder[index], NIL, index, 1);
                        }
                    }

                    // Determine if there is something in front of this comment.
                    before = (position > 0 && builder[position - 1] != LF);

                    // Determine if there is something behind this comment.
                    if (position + length < builder.Length)
                    {
                        behind = builder[position + length] != LF && !current.Scheme.EndOfLine;
                    }

                    // If there is nothing in front and nothing behind 
                    // this comment then it is a "pure" comment!
                    if (!before && !behind)
                    {
                        current.PureCommentCount++;
                        current.PureCommentLines += lines + 1; // Don't forget to add last line feed!
                    }
                    else
                    {
                        // In order to get a proper number of lines that only contain comments 
                        // it is necessary to recheck the "something in front" condition!
                        if (before)
                        {
                            for (int index = position - 1; index >= 0; index--)
                            {
                                // If only whitespaces exist between two comments 
                                // then it results in a countable comment line!
                                if (builder[index] != NIL && builder[index] != SPC && builder[index] != TAB)
                                {
                                    break; // Found at least one non-whitespace character.
                                }

                                if (index == 0 || builder[index - 1] == LF)
                                {
                                    before = !behind && builder[index] != NIL;
                                    lines++;
                                    break;
                                }
                            }
                        }

                        // Add an additional line if necessary.
                        if (!behind) { lines++; }

                        current.CodeCommentCount++;
                        current.CodeCommentLines += Math.Max(lines - 1, 0);
                    }

                    // Move to last comment character.
                    position += length - 1;
                    break;
                }
            }

            // Move to next character.
            return position += 1;
        }
    }
}
