﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.Collections.Generic;

namespace plexdata.LineCounter
{
    public class FileFilterGroup : plexdata.Utilities.NamedObject, ICloneable
    {
        public FileFilterGroup()
            : base()
        {
            this.Initialize();
        }

        public FileFilterGroup(string name)
            : base(name)
        {
            this.Initialize();
        }

        public FileFilterGroup(string name, string description)
            : base(name, description)
        {
            this.Initialize();
        }

        public FileFilterGroup(string name, string description, string tooltip)
            : base(name, description, tooltip)
        {
            this.Initialize();
        }

        public string DefaultTooltip
        {
            get
            {
                string result = this.Name.Trim();
                string pattern = this.SearchPattern.ToString().Trim();

                if (!String.IsNullOrEmpty(result))
                {
                    if (!String.IsNullOrEmpty(pattern))
                    {
                        result += " (" + pattern + ")";
                    }
                }
                return result;
            }
        }

        public SearchPattern SearchPattern { get; set; }

        public CommentSchemeCollection CommentSchemes { get; set; }

        public LiteralScheme StringLiteral { get; set; }

        public LiteralScheme CharacterLiteral { get; set; }

        public object Clone()
        {
            FileFilterGroup clone = new FileFilterGroup();

            clone.Name = this.Name.Clone() as string;
            clone.Description = this.Description.Clone() as string;
            clone.Tooltip = this.Tooltip.Clone() as string;

            if (this.Tag != null)
            {
                if (this.Tag is ICloneable)
                {
                    clone.Tag = (this.Tag as ICloneable).Clone();
                }
                else
                {
                    clone.Tag = this.Tag;
                }
            }
            else
            {
                clone.Tag = null;
            }

            if (this.SearchPattern != null)
            {
                clone.SearchPattern = this.SearchPattern.Clone() as SearchPattern;
            }
            else
            {
                clone.SearchPattern = null;
            }

            if (this.StringLiteral != null)
            {
                clone.StringLiteral = this.StringLiteral.Clone() as LiteralScheme;
            }
            else
            {
                clone.StringLiteral = null;
            }

            if (this.CharacterLiteral != null)
            {
                clone.CharacterLiteral = this.CharacterLiteral.Clone() as LiteralScheme;
            }
            else
            {
                clone.CharacterLiteral = null;
            }

            if (this.CommentSchemes != null)
            {
                clone.CommentSchemes = this.CommentSchemes.Clone() as CommentSchemeCollection;
            }
            else
            {
                clone.CommentSchemes = null;
            }

            return clone;
        }

        public override string ToString()
        {
            return this.Name;
        }

        private void Initialize()
        {
            this.SearchPattern = new SearchPattern();
            this.CommentSchemes = new CommentSchemeCollection();
            this.StringLiteral = new LiteralScheme();
            this.CharacterLiteral = new LiteralScheme();
        }
    }

    public class FileFilterGroupCollection : List<FileFilterGroup>, ICloneable
    {
        public FileFilterGroupCollection()
            : base()
        {
        }

        public void Sort(bool ascending)
        {
            base.Sort();
            if (!ascending)
            {
                base.Reverse();
            }
        }

        public object Clone()
        {
            FileFilterGroupCollection clone = new FileFilterGroupCollection();
            for (int index = 0; index < this.Count; index++)
            {
                FileFilterGroup current = this[index].Clone() as FileFilterGroup;
                if (current != null)
                {
                    clone.Add(current);
                }
            }
            return clone;
        }

        public static FileFilterGroupCollection DefaultSetup
        {
            get
            {
                FileFilterGroupCollection result = new FileFilterGroupCollection();

                FileFilterGroup group = new FileFilterGroup("C/C++ Files", "Standard group to filter C/C++ files.");
                group.SearchPattern = new SearchPattern(".c .cpp .h .hpp");
                group.StringLiteral = new LiteralScheme("\"", "\\");
                group.CharacterLiteral = new LiteralScheme("'", "\\");
                group.CommentSchemes = new CommentSchemeCollection();
                group.CommentSchemes.AddRange(
                    new CommentScheme[] { 
                        new CommentScheme("//"), 
                        new CommentScheme("///"), 
                        new CommentScheme("/*", "*/"), 
                        new CommentScheme("/**", "*/"), });
                result.Add(group);

                group = new FileFilterGroup("C# Files", "Standard group to filter C# files.");
                group.SearchPattern = new SearchPattern(".cs");
                group.StringLiteral = new LiteralScheme("\"", "\\");
                group.CharacterLiteral = new LiteralScheme("'", "\\");
                group.CommentSchemes = new CommentSchemeCollection();
                group.CommentSchemes.AddRange(
                    new CommentScheme[] { 
                        new CommentScheme("//"), 
                        new CommentScheme("///"), 
                        new CommentScheme("/*", "*/"), 
                        new CommentScheme("/**", "*/"), });
                result.Add(group);

                group = new FileFilterGroup("Java Source Files", "Standard group to filter Java files.");
                group.SearchPattern = new SearchPattern(".java");
                group.StringLiteral = new LiteralScheme("\"", "\\");
                group.CharacterLiteral = new LiteralScheme("'", "\\");
                group.CommentSchemes = new CommentSchemeCollection();
                group.CommentSchemes.AddRange(
                    new CommentScheme[] { 
                        new CommentScheme("//"), 
                        new CommentScheme("///"), 
                        new CommentScheme("/*", "*/"), });
                result.Add(group);

                group = new FileFilterGroup("Markup Files", "Standard group to filter HTML and XML files.");
                group.SearchPattern = new SearchPattern(".htm .html .xml");
                group.CommentSchemes = new CommentSchemeCollection();
                group.CommentSchemes.Add(new CommentScheme("<!--", "-->"));
                result.Add(group);

                group = new FileFilterGroup("Windows Batch Files", "Standard group to filter BAT and CMD files.");
                group.SearchPattern = new SearchPattern(".bat .cmd");
                group.CommentSchemes = new CommentSchemeCollection();
                group.CommentSchemes.Add(new CommentScheme("REM", true));
                result.Add(group);

                return result;
            }
        }
    }
}
