﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;
using System.ComponentModel;

namespace plexdata.LineCounter
{
    public partial class FileFilterGroupDialog : Form
    {
        ListViewItem newItem = null;
        ListViewItem editItem = null;
        ListViewItem copyItem = null;
        int errPadding = 0;
        string errWarning = String.Empty;

        public FileFilterGroupDialog()
            : base()
        {
            this.InitializeComponent();
            this.Icon = Icon.FromHandle(global::plexdata.LineCounter.Properties.Resources.FilterIcon.GetHicon());

            this.Load += this.OnDialogFormLoad;
            this.FormClosed += this.OnDialogFormClosed;

            this.errProvider.Icon = Icon.FromHandle(global::plexdata.LineCounter.Properties.Resources.WarningIcon.GetHicon());

            this.errPadding = -(this.errProvider.Icon.Width + 3);
            this.errWarning = "Leaving this field blank could cause a mismatch in counting literals!";
        }

        public FileFilterGroupDialog(FileFilterGroupCollection groups)
            : this()
        {
            this.SetupListItems(groups);

            try
            {
                // Try change comment box's font to a better one.
                Font font = new Font("Consolas", this.lstSchemes.Font.Size);
                if (font.Name != SystemFonts.DefaultFont.Name)
                {
                    this.lstSchemes.Font = font;
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        public FileFilterGroupCollection Groups { get; private set; }

        #region Dialog form event handlers.

        private void OnDialogFormLoad(object sender, EventArgs args)
        {
            // Don't use FormWindowState for a dialog window!

            // Set size before location because location depends on size!
            this.Size = Settings.Instance.FileFilterGroupWindow.Size;
            this.Location = Settings.Instance.FileFilterGroupWindow.Location;
        }

        private void OnDialogFormClosed(object sender, FormClosedEventArgs args)
        {
            // Don't use FormWindowState for a dialog window!

            // Set size before location because location depends on size!
            Settings.Instance.FileFilterGroupWindow.Size = this.Size;
            Settings.Instance.FileFilterGroupWindow.Location = this.Location;
        }

        #endregion // Dialog form event handlers.

        #region Private helper functions.

        private void SetupListItems(FileFilterGroupCollection groups)
        {
            // Setup internal File Filter list.
            if (groups != null)
            {
                this.Groups = groups.Clone() as FileFilterGroupCollection;
            }
            else
            {
                this.Groups = new FileFilterGroupCollection();
            }

            // Initialize groups list items.
            foreach (FileFilterGroup current in this.Groups)
            {
                ListViewItem item = new ListViewItem(current.Name);
                item.Tag = current;
                this.lstGroups.Items.Add(item);
            }

            this.AdjustColumnWidth(null);

            if (this.lstGroups.Items.Count > 0)
            {
                this.lstGroups.Items[0].Selected = true;
            }
        }

        private void AddNewListItem()
        {
            this.newItem = new ListViewItem("");
            this.newItem.Tag = new FileFilterGroup();
            this.lstGroups.Items.Add(this.newItem);
            this.newItem.BeginEdit();
        }

        private void EditSelectedListItem()
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                this.lstGroups.SelectedItems[0].BeginEdit();
            }
        }

        private void DeleteSelectedListItems()
        {
            int index = -1;

            foreach (ListViewItem selected in this.lstGroups.SelectedItems)
            {
                index = selected.Index;
                selected.Remove();
            }

            if (index >= 0)
            {
                if (index < this.lstGroups.Items.Count)
                {
                    this.lstGroups.Items[index].Selected = true;
                }
                else if (this.lstGroups.Items.Count > 0)
                {
                    this.lstGroups.Items[this.lstGroups.Items.Count - 1].Selected = true;
                }
            }

            this.AdjustColumnWidth(null);
        }

        private void AdjustColumnWidth(string label)
        {
            int width = 0;
            try
            {
                using (Graphics graphics = Graphics.FromHwnd(this.lstGroups.Handle))
                {
                    if (!String.IsNullOrEmpty(label))
                    {
                        width = graphics.MeasureString(label, this.lstGroups.Font).ToSize().Width + 10;
                    }
                    else
                    {
                        width = this.lstGroups.Width - 4;
                    }

                    foreach (ListViewItem current in this.lstGroups.Items)
                    {
                        if (current == this.editItem) { continue; }
                        int temp = graphics.MeasureString(current.Text, this.lstGroups.Font).ToSize().Width;
                        if (width < temp) { width = temp; }
                    }
                    this.lstGroups.Columns[0].Width = width;
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        private void SetStatusText()
        {
            this.SetStatusText(String.Empty);
        }

        private void SetStatusText(string tooltip)
        {
            this.sblTooltip.Text = (tooltip != null) ? tooltip : String.Empty;
        }

        private void SetStatusText(Control control)
        {
            if (control != null)
            {
                this.SetStatusText(this.tipDialog.GetToolTip(control));
            }
            else
            {
                this.SetStatusText(String.Empty);
            }
        }

        private void CheckStringEscape()
        {
            if (!String.IsNullOrEmpty(this.txtStringMarker.Text.Trim()) &&
                String.IsNullOrEmpty(this.txtStringEscape.Text.Trim()))
            {
                this.errProvider.SetIconAlignment(this.txtStringEscape, ErrorIconAlignment.MiddleRight);
                this.errProvider.SetIconPadding(this.txtStringEscape, this.errPadding);
                this.errProvider.SetError(this.txtStringEscape, this.errWarning);
            }
            else
            {
                this.errProvider.SetError(this.txtStringEscape, String.Empty);
            }
        }

        private void CheckCharacterEscape()
        {
            if (!String.IsNullOrEmpty(this.txtCharacterMarker.Text.Trim()) &&
                String.IsNullOrEmpty(this.txtCharacterEscape.Text.Trim()))
            {
                this.errProvider.SetIconAlignment(this.txtCharacterEscape, ErrorIconAlignment.MiddleRight);
                this.errProvider.SetIconPadding(this.txtCharacterEscape, this.errPadding);
                this.errProvider.SetError(this.txtCharacterEscape, this.errWarning);
            }
            else
            {
                this.errProvider.SetError(this.txtCharacterEscape, String.Empty);
            }
        }

        #endregion // Private helper functions.

        #region General button events.

        private void OnCancelButtonClick(object sender, EventArgs args)
        {
            this.DialogResult = DialogResult.Cancel;
            this.Close();
        }

        private void OnOkButtonClick(object sender, EventArgs args)
        {
            FileFilterGroupCollection groups = new FileFilterGroupCollection();
            foreach (ListViewItem current in this.lstGroups.Items)
            {
                FileFilterGroup group = current.Tag as FileFilterGroup;
                if (group != null)
                {
                    groups.Add(group);
                }
            }
            this.Groups = groups;
            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        #endregion // General button events.

        #region Groups list dependent events.

        private void OnGroupsListAfterLabelEdit(object sender, LabelEditEventArgs args)
        {
            if (!String.IsNullOrEmpty(args.Label))
            {
                FileFilterGroup group = this.lstGroups.Items[args.Item].Tag as FileFilterGroup;
                if (group != null)
                {
                    group.Name = args.Label;
                    this.AdjustColumnWidth(args.Label);
                    this.editItem = null;
                }
            }
            else
            {
                args.CancelEdit = true;
                if (this.newItem == this.lstGroups.Items[args.Item])
                {
                    this.lstGroups.Items.Remove(this.newItem);
                }
            }

            this.newItem = null;
        }

        private void OnGroupsListBeforeLabelEdit(object sender, LabelEditEventArgs args)
        {
            this.editItem = this.lstGroups.Items[args.Item];
        }

        private void OnGroupsListKeyDown(object sender, KeyEventArgs args)
        {
            if (args.KeyData == Keys.Return || args.KeyData == Keys.F2)
            {
                this.EditSelectedListItem();
            }
            else if (args.KeyData == Keys.Delete)
            {
                this.DeleteSelectedListItems();
            }
            else if (args.KeyData == Keys.Insert)
            {
                this.AddNewListItem();
            }
        }

        private void OnGroupsListSelectedIndexChanged(object sender, EventArgs args)
        {
            bool enabled = this.lstGroups.SelectedItems.Count > 0;

            this.btnGroupsEdit.Enabled = enabled;
            this.btnGroupsDelete.Enabled = enabled;

            this.btnSchemesAdd.Enabled = enabled;
            this.btnSchemesEdit.Enabled = enabled && this.lstSchemes.SelectedItems.Count > 0;
            this.btnSchemesDelete.Enabled = enabled && this.lstSchemes.SelectedItems.Count > 0;

            this.txtDescription.Enabled = enabled;
            this.txtTooltip.Enabled = enabled;
            this.txtPattern.Enabled = enabled;
            this.lstSchemes.Enabled = enabled;
            this.txtStringMarker.Enabled = enabled;
            this.txtStringEscape.Enabled = enabled;
            this.txtCharacterMarker.Enabled = enabled;
            this.txtCharacterEscape.Enabled = enabled;

            this.txtDescription.Text = String.Empty;
            this.txtTooltip.Text = String.Empty;
            this.txtPattern.Text = String.Empty;
            this.txtStringMarker.Text = String.Empty;
            this.txtStringEscape.Text = String.Empty;
            this.txtCharacterMarker.Text = String.Empty;
            this.txtCharacterEscape.Text = String.Empty;
            this.lstSchemes.Items.Clear();

            if (enabled)
            {
                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                if (group != null)
                {
                    this.txtDescription.Text = group.Description;
                    this.txtTooltip.Text = group.Tooltip;
                    this.txtPattern.Text = group.SearchPattern.ToString();
                    this.txtStringMarker.Text = group.StringLiteral.Marker;
                    this.txtStringEscape.Text = group.StringLiteral.Escape;
                    this.txtCharacterMarker.Text = group.CharacterLiteral.Marker;
                    this.txtCharacterEscape.Text = group.CharacterLiteral.Escape;
                    this.lstSchemes.Items.AddRange(group.CommentSchemes.ToArray());
                }
            }
        }

        private void OnGroupsListDoubleClick(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                this.lstGroups.SelectedItems[0].BeginEdit();
            }
        }

        private void OnGroupsListEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.lstGroups);
        }

        private void OnGroupsListLeave(object sender, EventArgs args)
        {
            this.SetStatusText();
        }

        private void OnGroupsAddButtonClick(object sender, EventArgs args)
        {
            this.AddNewListItem();
        }

        private void OnGroupsEditButtonClick(object sender, EventArgs args)
        {
            this.EditSelectedListItem();
        }

        private void OnGroupsDeleteButtonClick(object sender, EventArgs args)
        {
            this.DeleteSelectedListItems();
        }

        private void OnContextMenuGroupsOpening(object sender, CancelEventArgs args)
        {
            bool enabled = this.lstGroups.SelectedItems.Count > 0;
            this.cmiGroupsAdd.Enabled = true;
            this.cmiGroupsEdit.Enabled = enabled;
            this.cmiGroupsDelete.Enabled = enabled;
            this.cmiGroupsCopy.Enabled = enabled;
            this.cmiGroupsCut.Enabled = enabled;
            this.cmiGroupsPaste.Enabled = (this.copyItem != null);
        }

        private void OnContextMenuItemGroupsAddClick(object sender, EventArgs args)
        {
            this.AddNewListItem();
        }

        private void OnContextMenuItemGroupsEditClick(object sender, EventArgs args)
        {
            this.EditSelectedListItem();
        }

        private void OnContextMenuItemGroupsDeleteClick(object sender, EventArgs argse)
        {
            this.DeleteSelectedListItems();
        }

        private void OnContextMenuItemGroupsCopyClick(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                this.copyItem = (ListViewItem)(this.lstGroups.SelectedItems[0].Clone());
            }
        }

        private void OnContextMenuItemGroupsCutClick(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                this.copyItem = this.lstGroups.SelectedItems[0];
                this.copyItem.Remove();
            }
        }

        private void OnContextMenuItemGroupsPasteClick(object sender, EventArgs args)
        {
            if (this.copyItem != null)
            {
                if (this.lstGroups.SelectedItems.Count > 0)
                {
                    this.lstGroups.Items.Insert(this.lstGroups.SelectedItems[0].Index, this.copyItem);
                }
                else
                {
                    this.lstGroups.Items.Add(this.copyItem);
                }
                this.copyItem.Focused = true;
                this.copyItem.Selected = true;
            }
            this.copyItem = null;
        }

        private void OnContextMenuItemGroupsDefaultClick(object sender, EventArgs args)
        {
            if (this.lstGroups.Items.Count > 0)
            {
                DialogResult result = MessageBox.Show(
                    "Do you really want to reset current content\nwith application default values?",
                    "Restore Defaults", MessageBoxButtons.YesNo, MessageBoxIcon.Warning);
                if (result == DialogResult.No) { return; }

                this.lstGroups.Items.Clear();
            }

            this.SetupListItems(FileFilterGroupCollection.DefaultSetup);
        }

        #endregion // Groups list dependent events.

        #region Edit box events.

        private void OnEditDescriptionEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.txtDescription);
        }

        private void OnEditDescriptionLeave(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                if (group != null)
                {
                    group.Description = this.txtDescription.Text;
                    this.txtDescription.Text = group.Description;
                }
            }
            this.SetStatusText();
        }

        private void OnEditTooltipEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.txtTooltip);
        }

        private void OnEditTooltipLeave(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                if (group != null)
                {
                    group.Tooltip = this.txtTooltip.Text;
                    this.txtTooltip.Text = group.Tooltip;
                }
            }
            this.SetStatusText();
        }

        private void OnEditPatternEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.txtPattern);
        }

        private void OnEditPatternLeave(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                if (group != null)
                {
                    group.SearchPattern.Parse(this.txtPattern.Text);
                    this.txtPattern.Text = group.SearchPattern.ToString();
                }
            }
            this.SetStatusText();
        }

        private void OnEditCommentEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.lstSchemes);
        }

        private void OnEditStringMarkerEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.txtStringMarker);
        }

        private void OnEditStringMarkerLeave(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                if (group != null)
                {
                    group.StringLiteral.Marker = this.txtStringMarker.Text;
                    this.txtStringMarker.Text = group.StringLiteral.Marker;
                }
            }
            this.SetStatusText();
        }

        private void OnEditStringMarkerTextChanged(object sender, EventArgs args)
        {
            this.CheckStringEscape();
        }

        private void OnEditStringEscapeEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.txtStringEscape);
        }

        private void OnEditStringEscapeLeave(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                if (group != null)
                {
                    group.StringLiteral.Escape = this.txtStringEscape.Text;
                    this.txtStringEscape.Text = group.StringLiteral.Escape;
                }
            }
            this.SetStatusText();
        }

        private void OnEditStringEscapeTextChanged(object sender, EventArgs args)
        {
            this.CheckStringEscape();
        }

        private void OnEditStringEscapeValidated(object sender, EventArgs args)
        {
            this.CheckStringEscape();
        }

        private void OnEditCharacterMarkerEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.txtCharacterMarker);
        }

        private void OnEditCharacterMarkerLeave(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                if (group != null)
                {
                    group.CharacterLiteral.Marker = this.txtCharacterMarker.Text;
                    this.txtCharacterMarker.Text = group.CharacterLiteral.Marker;
                }
            }
            this.SetStatusText();
        }

        private void OnEditCharacterMarkerTextChanged(object sender, EventArgs args)
        {
            this.CheckCharacterEscape();
        }

        private void OnEditCharacterEscapeEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.txtCharacterEscape);
        }

        private void OnEditCharacterEscapeLeave(object sender, EventArgs args)
        {
            if (this.lstGroups.SelectedItems.Count > 0)
            {
                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                if (group != null)
                {
                    group.CharacterLiteral.Escape = this.txtCharacterEscape.Text;
                    this.txtCharacterEscape.Text = group.CharacterLiteral.Escape;
                }
            }
            this.SetStatusText();
        }

        private void OnEditCharacterEscapeTextChanged(object sender, EventArgs args)
        {
            this.CheckCharacterEscape();
        }

        private void OnEditCharacterEscapeValidated(object sender, EventArgs args)
        {
            this.CheckCharacterEscape();
        }

        #endregion // Edit box events.

        #region Schemes list dependent events.

        private void OnSchemesListEnter(object sender, EventArgs args)
        {
            this.SetStatusText(this.lstSchemes);
        }

        private void OnSchemesListLeave(object sender, EventArgs args)
        {
            this.SetStatusText();
        }

        private void OnSchemesSelectedIndexChanged(object sender, EventArgs args)
        {
            bool enabled = this.lstSchemes.SelectedItems.Count > 0;
            this.btnSchemesEdit.Enabled = enabled;
            this.btnSchemesDelete.Enabled = enabled;
        }

        private void OnSchemesListKeyDown(object sender, KeyEventArgs args)
        {
            if (args.KeyData == Keys.Insert)
            {
                this.btnSchemesAdd.PerformClick();
            }
            else if (args.KeyData == Keys.Return)
            {
                this.btnSchemesEdit.PerformClick();
            }
            else if (args.KeyData == Keys.Delete)
            {
                this.btnSchemesDelete.PerformClick();
            }
        }

        private void OnSchemesAddButtonClick(object sender, EventArgs args)
        {
            CommentSchemeDialog dialog = new CommentSchemeDialog();
            if (DialogResult.OK == dialog.ShowDialog(this))
            {
                if (this.lstGroups.SelectedItems.Count > 0)
                {
                    FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                    if (group != null)
                    {
                        group.CommentSchemes.Add(dialog.CommentScheme);
                        this.lstSchemes.Items.Clear();
                        this.lstSchemes.Items.AddRange(group.CommentSchemes.ToArray());
                    }
                }
            }
        }

        private void OnSchemesEditButtonClick(object sender, EventArgs args)
        {
            try
            {
                if (this.lstSchemes.SelectedIndex != -1)
                {
                    CommentSchemeDialog dialog = new CommentSchemeDialog(
                        this.lstSchemes.Items[this.lstSchemes.SelectedIndex] as CommentScheme);

                    if (DialogResult.OK == dialog.ShowDialog(this))
                    {
                        // Avoid duplicates!
                        if (!this.lstSchemes.Items.Contains(dialog.CommentScheme))
                        {
                            if (this.lstGroups.SelectedItems.Count > 0)
                            {
                                FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                                if (group != null)
                                {
                                    this.lstSchemes.Items[this.lstSchemes.SelectedIndex] = dialog.CommentScheme;
                                    group.CommentSchemes[this.lstSchemes.SelectedIndex] = dialog.CommentScheme;
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        private void OnSchemesDeleteButtonClick(object sender, EventArgs args)
        {
            try
            {
                if (this.lstSchemes.SelectedIndex != -1)
                {
                    FileFilterGroup group = this.lstGroups.SelectedItems[0].Tag as FileFilterGroup;
                    if (group != null)
                    {
                        int index = this.lstSchemes.SelectedIndex;
                        this.lstSchemes.Items.RemoveAt(index);
                        group.CommentSchemes.RemoveAt(index);
                    }
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        private void OnSchemesMouseDoubleClick(object sender, MouseEventArgs args)
        {
            this.btnSchemesEdit.PerformClick();
        }

        private void OnContextMenuSchemesOpening(object sender, CancelEventArgs args)
        {
            bool enabled = this.lstSchemes.SelectedItems.Count > 0;
            this.cmiSchemesAdd.Enabled = true;
            this.cmiSchemesEdit.Enabled = enabled;
            this.cmiSchemesDelete.Enabled = enabled;
        }

        private void OnContextMenuItemSchemesAddClick(object sender, EventArgs args)
        {
            this.btnSchemesAdd.PerformClick();
        }

        private void OnContextMenuItemSchemesEditClick(object sender, EventArgs args)
        {
            this.btnSchemesEdit.PerformClick();
        }

        private void OnContextMenuItemSchemesDeleteClick(object sender, EventArgs args)
        {
            this.btnSchemesDelete.PerformClick();
        }

        #endregion // Schemes list dependent events.
    }
}
