﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Linq;
using System.Threading;
using System.Diagnostics;
using System.Collections.Generic;

namespace plexdata.LineCounter
{
    public class FileScanCompletedArgs : EventArgs
    {
        public FileScanCompletedArgs(string filename)
            : base()
        {
            this.Filename = String.IsNullOrEmpty(filename) ? String.Empty : filename;
        }

        public string Filename { get; private set; }
    }

    public class ScanningStartedArgs : EventArgs
    {
        public ScanningStartedArgs(int count)
            : base()
        {
            this.Count = count;
        }

        public int Count { get; private set; }
    }

    public class FileScanner
    {
        public event EventHandler<ScanningStartedArgs> ScanningStarted;
        public event EventHandler<EventArgs> ScanningCompleted;
        public event EventHandler<FileScanCompletedArgs> FileScanCompleted;

        private List<ContentScanner> scanners = null;
        private Thread thread = null;

        public FileScanner(FileFilterGroup filter)
            : base()
        {
            if (filter != null)
            {
                // Save given file group filter.
                this.Filter = filter;

                // Initialize internal management lists.
                this.scanners = new List<ContentScanner>();
            }
            else
            {
                throw new ArgumentNullException();
            }
        }

        public FileFilterGroup Filter { get; private set; }

        public ContentScanner[] ScannersByExtension(string extension)
        {
            List<ContentScanner> result = new List<ContentScanner>();
            try
            {
                result = new List<ContentScanner>(this.scanners.Where(
                    item => String.Compare(extension, Path.GetExtension(item.Filename), true) == 0));
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            return result.ToArray();
        }

        // Scans all files recursively.
        public bool ScanAllFolders(string baseFolder)
        {
            bool success = false;
            if (!String.IsNullOrEmpty(baseFolder))
            {
                List<string> folderList = new List<string>();

                // Find all folders recursively.
                this.BuildFolderList(baseFolder, ref folderList);
                if (folderList.Count > 0)
                {
                    success = this.ScanFolders(folderList.ToArray());
                }
            }
            return success;
        }

        // Scans all files only in given folder list.
        public bool ScanFolders(string[] folderList)
        {
            bool success = false;
            if (folderList != null && folderList.Length > 0)
            {
                try
                {
                    // TODO: Check if recursive file search (preventing duplicates!) is meaningful!

                    List<string> fileList = new List<string>();
                    List<string> patternList = new List<string>(this.Filter.SearchPattern.Values);
                    foreach (string folder in folderList)
                    {
                        // First find all matching files.
                        foreach (string pattern in patternList)
                        {
                            // Additional list filtering is necessary because 
                            // of the fact that function GetFiles() returns one 
                            // and the same file twice when the file pattern 
                            // "look" similar, like *.htm and *.html. Damn shit.
                            fileList.AddRange(Directory.GetFiles(folder, pattern).Where(
                                item => String.Compare(Path.GetExtension(item),
                                    Path.GetExtension(pattern), true) == 0));
                        }
                    }

                    // Find all folders recursively.
                    if (fileList.Count > 0)
                    {
                        success = this.ScanFiles(fileList.ToArray());
                    }
                }
                catch (Exception exception)
                {
                    Debug.WriteLine(exception);
                }
            }
            return success;
        }

        // Scans all given files.
        public bool ScanFiles(string[] fileList)
        {
            bool success = false;
            if (fileList != null && fileList.Length > 0)
            {
                try
                {
                    if (this.thread == null)
                    {
                        this.scanners.Clear();
                        foreach (string file in fileList)
                        {
                            this.scanners.Add(
                                new ContentScanner(file, this.Filter.CommentSchemes));
                        }

                        this.thread = new Thread(this.DoWork);
                        this.thread.Start();

                        while (!this.thread.IsAlive) { }

                        success = true;
                    }
                }
                catch (Exception exception)
                {
                    Debug.WriteLine(exception);
                }
            }
            return success;
        }

        public void AbortScanning()
        {
            try
            {
                if (this.thread != null && this.thread.IsAlive)
                {
                    this.thread.Abort();
                    this.thread.Join();
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            finally
            {
                this.thread = null;
            }
        }

        private void DoWork()
        {
            try
            {
                if (this.ScanningStarted != null)
                {
                    this.ScanningStarted(
                        this, new ScanningStartedArgs(this.scanners.Count));
                }

                for (int index = 0; index < this.scanners.Count; index++)
                {
                    this.scanners[index].Scan(this.Filter.StringLiteral, this.Filter.CharacterLiteral);

                    if (this.FileScanCompleted != null)
                    {
                        this.FileScanCompleted(
                            this, new FileScanCompletedArgs(this.scanners[index].Filename));
                    }
                }

                if (this.ScanningCompleted != null)
                {
                    this.ScanningCompleted(
                        this, EventArgs.Empty);
                }
            }
            catch (ThreadAbortException)
            {
                // Expected exception because of the thread is simply killed.
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        private void BuildFolderList(string parent, ref List<string> result)
        {
            try
            {
                if (!String.IsNullOrEmpty(parent))
                {
                    // First add parent folder.
                    result.Add(parent);

                    // Now repeat for all available child folders.
                    foreach (string current in Directory.GetDirectories(parent))
                    {
                        this.BuildFolderList(current, ref result);
                    }
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }
    }
}
