﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Drawing;
using System.Threading;
using System.Reflection;
using System.Collections;
using System.Diagnostics;
using System.Globalization;
using System.Windows.Forms;
using System.Collections.Generic;

namespace plexdata.LineCounter
{
    public abstract partial class ItemSelectorDialog : Form
    {
        // BUG: Class ItemSelectorDialog has a bug while selecting and sorting list items.
        //      Under some circumstances selecting only a few items followed by sorting the 
        //      list (not sure if both columns are affected or just column Parent) causes 
        //      other items to become selected too. A specific pattern couldn't be detected 
        //      (means: it wasn't the same range selected or something like that). 
        //      Unfortunately, this bug is not reproducible and cannot be fixed at the moment.

        #region Private helper class implementation.

        /// <summary>
        /// This class is an implementation of the 'IComparer' interface.
        /// </summary>
        /// <remarks>
        /// This class is simply copied from http://support.microsoft.com/kb/319401
        /// </remarks>
        public class RepositoryItemSorter : IComparer<ListViewItem>
        {
            /// <summary>
            /// Case insensitive comparer object
            /// </summary>
            private CaseInsensitiveComparer objectCompare;

            /// <summary>
            /// Class constructor.  Initializes various elements
            /// </summary>
            public RepositoryItemSorter()
            {
                // Initialize the column to '0'
                this.SortColumn = 0;

                // Initialize the sort order to 'none'
                this.SortOrder = SortOrder.None;

                // Initialize the CaseInsensitiveComparer object
                this.objectCompare = new CaseInsensitiveComparer();
            }

            /// <summary>
            /// This method is inherited from the IComparer interface. It 
            /// compares the two objects passed using a case insensitive 
            /// comparison.
            /// </summary>
            /// <param name="left">
            /// First object to be compared
            /// </param>
            /// <param name="right">
            /// Second object to be compared
            /// </param>
            /// <returns>
            /// The result of the comparison. "0" if equal, negative if 'left' 
            /// is less than 'right' and positive if 'left' is greater than 
            /// 'right'
            /// </returns>
            public int Compare(ListViewItem left, ListViewItem right)
            {
                // Compare the two items
                int result = this.objectCompare.Compare(
                    left.SubItems[this.SortColumn].Text,
                    right.SubItems[this.SortColumn].Text);

                // Calculate correct return value based on object comparison
                if (this.SortOrder == SortOrder.Ascending)
                {
                    // Ascending sort is selected, return normal result of compare operation
                    return result;
                }
                else if (this.SortOrder == SortOrder.Descending)
                {
                    // Descending sort is selected, return negative result of compare operation
                    return (-result);
                }
                else
                {
                    // Return '0' to indicate they are equal
                    return 0;
                }
            }

            /// <summary>
            /// Gets or sets the number of the column to which to apply 
            /// the sorting operation (Defaults to '0').
            /// </summary>
            public int SortColumn { get; set; }

            /// <summary>
            /// Gets or sets the order of sorting to apply (for example, 
            /// 'Ascending' or 'Descending').
            /// </summary>
            public SortOrder SortOrder { get; set; }
        }

        #endregion // Private helper class implementation.

        private List<ListViewItem> repository = null;
        private RepositoryItemSorter repositorySorter = null;
        private Thread worker = null;

        public ItemSelectorDialog()
            : base()
        {
            this.InitializeComponent();
            this.Icon = Icon.FromHandle(global::plexdata.LineCounter.Properties.Resources.IncludeIcon.GetHicon());

            this.BasePath = String.Empty;
            this.SelectedItems = new string[] { };

            this.repository = new List<ListViewItem>();
            this.repositorySorter = new RepositoryItemSorter();

            PropertyInfo property;
            property = typeof(ListView).GetProperty("DoubleBuffered", BindingFlags.NonPublic | BindingFlags.Instance);
            property.SetValue(this.lstItemView, true, null);

            this.lstItemView.Focus();
        }

        #region Abstract member declaration.

        protected abstract void ScanFolder(string parent);

        #endregion // Abstract member declaration.

        #region Public property definition.

        public string BasePath
        {
            get { return this.lblBasePathValue.Text; }
            set
            {
                if (value != this.lblBasePathValue.Text)
                {
                    this.lblBasePathValue.Text = value;
                }
            }
        }

        public string[] SelectedItems { get; protected set; }

        #endregion // Public property definition.

        #region Protected member definition.

        protected void AddListItems(string pathname)
        {
            this.AddListItems(new string[] { pathname });
        }

        private delegate void AddListItemsDelegate(string[] pathlist);
        protected void AddListItems(string[] pathlist)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new AddListItemsDelegate(AddListItems), new object[] { pathlist });
            }
            else
            {
                this.lstItemView.BeginUpdate();
                try
                {
                    // TODO: Check for duplicates...

                    if (pathlist != null)
                    {
                        foreach (string fullpath in pathlist)
                        {
                            string candidate = Path.GetFileName(fullpath);
                            string parent = Path.GetDirectoryName(fullpath);

                            ListViewItem item = new ListViewItem();
                            item.Text = candidate;
                            item.Tag = fullpath;
                            item.Checked = true; // Is there a bug in virtual list mode? If checked == false, then the checkbox is invisible.
                            item.Checked = this.IsChecked(fullpath);

                            try
                            {
                                FileInfo info = new FileInfo(fullpath);
                                if (info.Exists)
                                {
                                    if ((info.Attributes & FileAttributes.Directory) == FileAttributes.Directory)
                                    {
                                        item.ToolTipText = String.Format(
                                            "Number of Files: {0}\nNumber of Folders: {1}",
                                            this.NumberDisplayString(Directory.GetFiles(info.FullName).Length),
                                            this.NumberDisplayString(Directory.GetDirectories(info.FullName).Length));
                                    }
                                    else
                                    {
                                        item.ToolTipText = String.Format(
                                            "File Size: {0} Bytes\nCreated: {1} (UTC)\nChanged: {2} (UTC)",
                                            this.NumberDisplayString(info.Length),
                                            this.DateTimeDisplayString(info.CreationTimeUtc),
                                            this.DateTimeDisplayString(info.LastWriteTimeUtc));
                                    }
                                }
                            }
                            catch
                            {
                                item.ToolTipText = fullpath;
                            }

                            item.SubItems.Add(parent);
                            this.repository.Add(item);
                        }

                        this.lstItemView.VirtualListSize = this.repository.Count;

                        this.sblTotalValue.Text = this.NumberDisplayString(this.repository.Count);
                    }
                }
                catch (Exception exception)
                {
                    Debug.WriteLine(exception);
                }
                finally
                {
                    this.lstItemView.EndUpdate();
                }
            }
        }

        #endregion // Protected member definition.

        #region Private member definition.

        private bool IsChecked(string fullpath)
        {
            if (this.SelectedItems != null && this.SelectedItems.Length > 0)
            {
                foreach (string current in this.SelectedItems)
                {
                    if (String.Compare(current, fullpath, true) == 0)
                    {
                        return true;
                    }
                }
                return false; // Don't select item if not found.
            }
            return true; // No list, then select item by default.
        }

        private string NumberDisplayString(long value)
        {
            return value.ToString("N0", NumberFormatInfo.CurrentInfo);
        }

        private string DateTimeDisplayString(DateTime value)
        {
            return value.Year.ToString("0000") + "-" +
                   value.Month.ToString("00") + "-" +
                   value.Day.ToString("00") + " " +
                   value.ToLongTimeString();
        }

        private delegate void WorkerThreadStartedDelegate();
        private void WorkerThreadStarted()
        {
            try
            {
                if (this.InvokeRequired)
                {
                    this.Invoke(new WorkerThreadStartedDelegate(this.WorkerThreadStarted));
                }
                else
                {
                    this.Cursor = Cursors.AppStarting;
                    this.chkShowList.Enabled = false;
                    this.btnOk.Enabled = false;
                    this.btnCheckAll.Enabled = false;
                    this.btnUncheckAll.Enabled = false;
                    this.btnToggleSelected.Enabled = false;
                    this.lstItemView.HeaderStyle = ColumnHeaderStyle.Nonclickable;
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        private delegate void WorkerThreadEndedDelegate();
        private void WorkerThreadEnded()
        {
            try
            {
                if (this.InvokeRequired)
                {
                    this.Invoke(new WorkerThreadEndedDelegate(this.WorkerThreadEnded));
                }
                else
                {
                    this.Cursor = Cursors.Default;
                    this.chkShowList.Enabled = true;
                    this.btnOk.Enabled = true;
                    this.btnCheckAll.Enabled = true;
                    this.btnUncheckAll.Enabled = true;
                    this.btnToggleSelected.Enabled = true;
                    this.lstItemView.HeaderStyle = ColumnHeaderStyle.Clickable;
                    this.colItem.Width = -1;
                    this.colParent.Width = -2;
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        private void WorkerThread()
        {
            this.WorkerThreadStarted();
            try
            {
                this.ScanFolder(this.BasePath);
            }
            catch (ThreadAbortException)
            {
                // Expected!
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            finally
            {
                this.WorkerThreadEnded();
            }
        }

        #endregion // Private member definition.

        #region Event handler implementation.

        private void OnItemSelectorDialogShown(object sender, EventArgs args)
        {
            int width = this.lstItemView.DisplayRectangle.Width - SystemInformation.VerticalScrollBarWidth - 1;
            this.colItem.Width = (int)(width * 0.3);
            this.colParent.Width = (int)(width * 0.7);

            this.worker = new Thread(this.WorkerThread);
            this.worker.Start();
        }

        private void OnItemSelectorDialogClosing(object sender, FormClosingEventArgs args)
        {
            if (this.worker != null && this.worker.IsAlive)
            {
                try
                {
                    this.worker.Abort();
                }
                catch (Exception exception)
                {
                    Debug.WriteLine(exception);
                }
                finally
                {
                    this.worker = null;
                }
            }
        }

        private void OnCheckAllButtonClick(object sender, EventArgs args)
        {
            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.AppStarting;
            this.lstItemView.BeginUpdate();
            try
            {
                foreach (ListViewItem current in this.repository)
                {
                    current.Checked = true;
                }
            }
            finally
            {
                this.lstItemView.EndUpdate();
                this.Cursor = cursor;
            }
        }

        private void OnUncheckAllButtonClick(object sender, EventArgs args)
        {
            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.AppStarting;
            this.lstItemView.BeginUpdate();
            try
            {
                foreach (ListViewItem current in this.repository)
                {
                    current.Checked = false;
                }
            }
            finally
            {
                this.lstItemView.EndUpdate();
                this.Cursor = cursor;
            }
        }

        private void OnToggleSelectedButtonClick(object sender, EventArgs args)
        {
            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.AppStarting;
            this.lstItemView.BeginUpdate();
            try
            {
                if (this.lstItemView.SelectedIndices.Count > 1)
                {
                    // Only affect selected items.
                    foreach (int index in this.lstItemView.SelectedIndices)
                    {
                        if (0 <= index && index < this.repository.Count)
                        {
                            this.repository[index].Checked = !this.repository[index].Checked;
                        }
                    }
                }
                else
                {
                    // Otherwise affect all items.
                    foreach (ListViewItem current in this.repository)
                    {
                        current.Checked = !current.Checked;
                    }
                }
            }
            finally
            {
                this.lstItemView.EndUpdate();
                this.Cursor = cursor;
            }
        }

        private void OnCancelButtonClick(object sender, EventArgs args)
        {
            this.DialogResult = DialogResult.Cancel;
        }

        private void OnOkButtonClick(object sender, EventArgs args)
        {
            this.DialogResult = DialogResult.OK;
            List<string> helper = new List<string>();
            foreach (ListViewItem item in this.repository)
            {
                if (item.Checked && item.Tag is string)
                {
                    helper.Add(item.Tag as string);
                }
            }
            this.SelectedItems = helper.ToArray();
        }

        private void OnViewListCheckBoxCheckedChanged(object sender, EventArgs args)
        {
            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.AppStarting;
            this.lstItemView.BeginUpdate();
            try
            {
                this.lstItemView.View = this.chkShowList.Checked ? View.List : View.Details;
            }
            finally
            {
                this.lstItemView.EndUpdate();
                this.lstItemView.Invalidate();
                this.Cursor = cursor;
            }
        }

        private void OnListViewRetrieveVirtualItem(object sender, RetrieveVirtualItemEventArgs args)
        {
            int index = args.ItemIndex;
            if (0 <= index && index < this.repository.Count)
            {
                args.Item = this.repository[index];
            }
            else
            {
                args.Item = null;
            }
        }

        private void OnListViewMouseClick(object sender, MouseEventArgs args)
        {
            ListView view = (ListView)sender;
            ListViewHitTestInfo info = view.HitTest(args.Location);
            if (info.Location == ListViewHitTestLocations.StateImage)
            {
                int index = info.Item.Index;
                if (0 <= index && index < this.repository.Count)
                {
                    this.repository[index].Checked = !this.repository[index].Checked;
                    view.Invalidate(info.Item.Bounds);
                }
            }
        }

        private void OnListViewKeyDown(object sender, KeyEventArgs args)
        {
            // Do nothing until search has finished!
            if (this.worker != null && this.worker.IsAlive) { return; }

            if (this.lstItemView.SelectedIndices.Count > 0)
            {
                if (!args.Control && !args.Alt && !args.Shift)
                {
                    if (args.KeyCode == Keys.Space)
                    {
                        Cursor cursor = this.Cursor;
                        this.Cursor = Cursors.AppStarting;
                        this.lstItemView.BeginUpdate();
                        try
                        {
                            // Only affect selected items.
                            foreach (int index in this.lstItemView.SelectedIndices)
                            {
                                if (0 <= index && index < this.repository.Count)
                                {
                                    this.repository[index].Checked = !this.repository[index].Checked;
                                }
                            }
                        }
                        finally
                        {
                            this.lstItemView.EndUpdate();
                            this.Cursor = cursor;
                        }
                    }
                }
            }
        }

        private void OnListViewColumnClick(object sender, ColumnClickEventArgs args)
        {
            // Do nothing until search has finished!
            if (this.worker != null && this.worker.IsAlive) { return; }

            this.lstItemView.BeginUpdate();
            try
            {
                // Determine if clicked column is already the column that is being sorted.
                if (args.Column == this.repositorySorter.SortColumn)
                {
                    // Reverse the current sort direction for this column.
                    if (this.repositorySorter.SortOrder == SortOrder.Ascending)
                    {
                        this.repositorySorter.SortOrder = SortOrder.Descending;
                    }
                    else
                    {
                        this.repositorySorter.SortOrder = SortOrder.Ascending;
                    }
                }
                else
                {
                    // Set the column number that is to be sorted; default to ascending.
                    this.repositorySorter.SortColumn = args.Column;
                    this.repositorySorter.SortOrder = SortOrder.Ascending;
                }

                // Perform the sort with these new sort options.
                this.repository.Sort(this.repositorySorter);
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            finally
            {
                this.lstItemView.EndUpdate();
            }
        }

        #endregion // Event handler implementation.
    }
}
