﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Drawing;
using System.Diagnostics;
using System.Globalization;
using System.Windows.Forms;
using System.Collections.Generic;

using plexdata.Generic;

namespace plexdata.LineCounter
{
    public partial class MainForm : Form
    {
        private static string OPEN_SAVE_EXTENSION = ".env";
        private static string OPEN_SAVE_FILTER = "Line Counter file (*.env)|*.env|All files (*.*)|*.*";

        private static string REPORT_SAVE_EXTENSION = ".html";
        private static string REPORT_SAVE_FILTER = "Line Counter Report files (*.html)|*.html;*.htm|All files (*.*)|*.*";

        private FileScanner fileScanner = null;
        private ScanningResults scanningResults = null;

        public MainForm()
        {
            this.InitializeComponent();
            this.Icon = Icon.FromHandle(global::plexdata.LineCounter.Properties.Resources.MainIcon.GetHicon());

            // Start with a new environment.
            this.NewEnvironment();
            this.IsDirty = false;

            this.HideResults();
            this.ResetScanValues();
        }

        public LineCounterEnvironment Environment { get; private set; }

        private bool isDirty = false;
        public bool IsDirty
        {
            get { return this.isDirty; }
            private set
            {
                this.isDirty = value;
                // Do this in any case!
                this.UpdateControls();
            }
        }

        private bool IsValidBasePath
        {
            get
            {
                bool valid = false;
                try
                {
                    string basePath = this.BasePath;
                    if (!String.IsNullOrEmpty(basePath))
                    {
                        DirectoryInfo info = new DirectoryInfo(this.BasePath);
                        valid = info.Exists && ((info.Attributes & FileAttributes.Directory) == FileAttributes.Directory);
                    }
                }
                catch (Exception exception)
                {
                    Debug.WriteLine(exception);
                }
                return valid;
            }
        }

        private string BasePath
        {
            get { return this.txtBasePath.Text.Trim().Replace("\"", ""); }
            set
            {
                if (value == null)
                {
                    value = String.Empty;
                }
                this.txtBasePath.Text = value.Trim().Replace("\"", "");
            }
        }

        #region Private helper functions.

        private void ShowResults()
        {
            if (this.scanningResults != null)
            {
                this.tbcResults.Initialize(this.scanningResults);
            }
            else
            {
                this.tbcResults.Cleanup();
            }
        }

        private void HideResults()
        {
            this.tbcResults.Cleanup();
        }

        private void ResetScanValues()
        {
            this.prbExecution.Value = 0;
            this.sblFileValue.Text = String.Empty;
            this.sblScannedValue.Text = String.Empty;
        }

        private void SwapButtonLabel(Button button)
        {
            // Swap button label from belonging tag.
            if (button != null && button.Tag is string)
            {
                string label = button.Text;
                button.Text = button.Tag as string;
                button.Tag = label;
            }
        }

        private void UpdateControls()
        {
            this.tbbExit.Enabled = true;
            this.tbbNew.Enabled = true;
            this.tbbOpen.Enabled = true;
            this.tbbSave.Enabled = this.IsDirty;
            this.tbbInclude.Enabled = this.Environment.GroupCollection.Count > 0;
            this.tbbReport.Enabled = this.scanningResults != null;
            this.tbbSettings.Enabled = false; // RFU
            this.tbbAbout.Enabled = true;

            if (!this.tbbInclude.Enabled)
            {
                this.tipMain.SetToolTip(this.cmbFileGroups, String.Empty);
            }
        }

        private bool HandleDirty()
        {
            // Check if it is nothing to do.
            if (!this.IsDirty) { return true; }

            DialogResult result = MessageBox.Show(
                "The settings have been changed! Would you like to save made changes?",
                this.Text, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Warning);

            if (result == DialogResult.Cancel)
            {
                return false;
            }
            else
            {
                if (result == DialogResult.Yes)
                {
                    this.tbbSave.PerformClick();
                }
                return true;
            }
        }

        private void NewEnvironment()
        {
            this.Environment = new LineCounterEnvironment(FileFilterGroupCollection.DefaultSetup);
            this.AssignEnvironment();
            this.IsDirty = false;
        }

        private void OpenEnvironment(string filename)
        {
            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.WaitCursor;
            try
            {
                LineCounterEnvironment environment;
                if (Serializer.Load<LineCounterEnvironment>(filename, out environment))
                {
                    Settings.Instance.General.LastEnvironment = filename;
                    this.Environment = environment;
                    this.AssignEnvironment();
                    this.IsDirty = false;
                }
                else
                {
                    string message = String.Format("Could not open environment file {0}!", Path.GetFileName(filename));
                    if (Serializer.Exception != null)
                    {
                        message += "\n\nSystem reported: " + Serializer.Exception.Message;
                    }
                    MessageBox.Show(this, message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
            finally
            {
                this.Cursor = cursor;
            }
        }

        private void SaveEnvironment(string filename)
        {
            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.WaitCursor;
            try
            {
                this.AssignEnvironment(true);
                if (Serializer.Save<LineCounterEnvironment>(filename, this.Environment))
                {
                    Settings.Instance.General.LastEnvironment = filename;
                    this.IsDirty = false;
                }
                else
                {
                    string message = String.Format("Could not save environment file {0}!", Path.GetFileName(filename));
                    if (Serializer.Exception != null)
                    {
                        message += "\n\nSystem reported: " + Serializer.Exception.Message;
                    }
                    MessageBox.Show(this, message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
            finally
            {
                this.Cursor = cursor;
            }
        }

        private void SaveReport(string filename, ReportGenerator.ReportType reportType)
        {
            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.WaitCursor;
            try
            {
                Settings.Instance.General.LastReportFile = filename;

                string reportContent = ReportGenerator.GetExportReport(
                    this.scanningResults, reportType);

                if (reportContent.Length > 0)
                {
                    using (StreamWriter writer = File.CreateText(filename))
                    {
                        writer.Write(reportContent);
                    }
                }
            }
            finally
            {
                this.Cursor = cursor;
            }
        }

        private void AssignEnvironment()
        {
            this.AssignEnvironment(false);
        }

        private void AssignEnvironment(bool save)
        {
            // TODO: Update or save more settings!

            if (this.Environment != null)
            {
                if (save)
                {
                    // Save base path value.
                    this.Environment.BasePath = this.BasePath;

                    // Save folder mode value.
                    this.Environment.FolderMode = this.tbmIncludeSearchFolders.Checked;

                    // Save combo box last selected group name.
                    this.Environment.LastGroup = this.GetSelectedFileGroup();
                }
                else
                {
                    this.HideResults();
                    this.ResetScanValues();

                    // Set base path value.
                    this.BasePath = this.Environment.BasePath;

                    // Set folder mode value.
                    this.tbmIncludeSearchFolders.Checked = this.Environment.FolderMode;

                    // Set combo box entries and try select last one.
                    this.cmbFileGroups.DataSource = this.Environment.GroupCollection;
                    this.SetSelectedFileGroup(this.Environment.LastGroup);
                }
            }
        }

        private string GetSelectedFileGroup()
        {
            FileFilterGroup group = this.cmbFileGroups.SelectedItem as FileFilterGroup;
            return (group != null) ? group.Name : String.Empty;
        }

        private void SetSelectedFileGroup(FileFilterGroup group)
        {
            if (group != null && !String.IsNullOrEmpty(group.Name))
            {
                this.SetSelectedFileGroup(group.Name);
            }
        }

        private void SetSelectedFileGroup(string label)
        {
            if (!String.IsNullOrEmpty(label) && this.cmbFileGroups.Items.Count > 0)
            {
                foreach (object current in this.cmbFileGroups.Items)
                {
                    FileFilterGroup group = current as FileFilterGroup;
                    if (group != null)
                    {
                        if (group.CompareTo(label) == 0)
                        {
                            this.cmbFileGroups.SelectedItem = group;
                            return;
                        }
                    }
                }
                this.cmbFileGroups.SelectedItem = this.cmbFileGroups.Items[0];
            }
        }

        private string GetValidDirectory(string path)
        {
            string result = String.Empty;
            try
            {
                if (String.IsNullOrEmpty(path))
                {
                    path = Application.ExecutablePath;
                }

                FileInfo info = new FileInfo(path);
                if (info.Exists)
                {
                    if ((info.Attributes & FileAttributes.Directory) == FileAttributes.Directory)
                    {
                        result = info.FullName;
                    }
                    else
                    {
                        result = info.DirectoryName;
                    }
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            return result;
        }

        private string GetValidFilename(string path)
        {
            return this.GetValidFilename(path, MainForm.OPEN_SAVE_EXTENSION);
        }

        private string GetValidFilename(string path, string extension)
        {
            string result = String.Empty;
            try
            {
                if (String.IsNullOrEmpty(path))
                {
                    path = Path.GetFileNameWithoutExtension(Application.ExecutablePath) +
                           Path.GetExtension(extension);
                }

                FileInfo info = new FileInfo(path);
                if (!info.Exists || (info.Attributes & FileAttributes.Directory) != FileAttributes.Directory)
                {
                    result = Path.GetFileName(info.FullName);
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            return result;
        }

        private bool IsBasePathValidated(out string result)
        {
            bool success = false;

            result = this.BasePath;

            if (!String.IsNullOrEmpty(result))
            {
                DirectoryInfo info = new DirectoryInfo(result);
                if (info.Attributes != (FileAttributes)(-1))
                {
                    if ((info.Attributes & FileAttributes.Directory) == FileAttributes.Directory)
                    {
                        success = true;
                    }
                    else
                    {
                        MessageBox.Show(this,
                            "The path given as base path is a file. Please " +
                            "provide just a directory name as base path.",
                            this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
                else
                {
                    MessageBox.Show(this,
                        "The path given as base path is invalid. Please " +
                        "provide an existing directory name as base path.",
                        this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
            return success;
        }

        #endregion // Private helper functions.

        #region Main form control event handlers.

        private void OnBasePathButtonClick(object sender, EventArgs args)
        {
            try
            {
                FolderBrowserDialog browser = new FolderBrowserDialog();
                browser.Description = "Choose the base folder where searching operation should begin from.";
                browser.SelectedPath = this.BasePath;
                browser.ShowNewFolderButton = false;

                if (DialogResult.OK == browser.ShowDialog(this))
                {
                    this.BasePath = browser.SelectedPath;
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        private void OnBasePathTextChanged(object sender, EventArgs args)
        {
            this.IsDirty = true;
        }

        private void OnFileGroupsSelectedIndexChanged(object sender, EventArgs args)
        {
            FileFilterGroup group = this.cmbFileGroups.SelectedItem as FileFilterGroup;
            if (group != null)
            {
                string tooltip = group.Tooltip.Trim();
                if (String.IsNullOrEmpty(tooltip))
                {
                    tooltip = group.DefaultTooltip;
                }
                this.tipMain.SetToolTip(this.cmbFileGroups, tooltip);
            }
            else
            {
                this.tipMain.SetToolTip(this.cmbFileGroups, String.Empty);
            }
            this.IsDirty = true;
        }

        private void OnExecuteButtonClick(object sender, EventArgs args)
        {
            try
            {
                if (this.fileScanner == null)
                {
                    this.Cursor = Cursors.WaitCursor;

                    this.HideResults();
                    this.ResetScanValues();
                    this.btnExecute.Enabled = false;

                    this.fileScanner = new FileScanner(this.cmbFileGroups.SelectedItem as FileFilterGroup);
                    this.fileScanner.ScanningStarted += this.OnScanningStarted;
                    this.fileScanner.FileScanCompleted += this.OnFileScanCompleted;
                    this.fileScanner.ScanningCompleted += this.OnScanningCompleted;

                    bool success = false;

                    // Both lists are set, let the checkbox decide.
                    if (this.Environment.IsSelectedFiles && this.Environment.IsSelectedFolders)
                    {
                        if (this.tbmIncludeSearchFolders.Checked)
                        {
                            success = this.fileScanner.ScanFolders(this.Environment.SelectedFolders);
                        }
                        else
                        {
                            success = this.fileScanner.ScanFiles(this.Environment.SelectedFiles);
                        }
                    }
                    // Only folder list is set, checkbox not needed.
                    else if (!this.Environment.IsSelectedFiles && this.Environment.IsSelectedFolders)
                    {
                        success = this.fileScanner.ScanFolders(this.Environment.SelectedFolders);
                    }
                    // Only file list is set, checkbox not needed.
                    else if (this.Environment.IsSelectedFiles && !this.Environment.IsSelectedFolders)
                    {
                        success = this.fileScanner.ScanFiles(this.Environment.SelectedFiles);
                    }
                    // Neither file list nor folder list is set, start an overall search.
                    else
                    {
                        string baseBath;
                        if (this.IsBasePathValidated(out baseBath))
                        {
                            success = this.fileScanner.ScanAllFolders(baseBath);
                        }
                    }

                    if (!success)
                    {
                        this.fileScanner.AbortScanning();
                        this.fileScanner.ScanningStarted -= this.OnScanningStarted;
                        this.fileScanner.FileScanCompleted -= this.OnFileScanCompleted;
                        this.fileScanner.ScanningCompleted -= this.OnScanningCompleted;
                        this.fileScanner = null;

                        this.sblFileValue.Text = "Failed!";
                        this.btnExecute.Enabled = true;
                        this.Cursor = Cursors.Default;
                    }
                    else
                    {
                        this.SwapButtonLabel(this.btnExecute);
                        this.Cursor = Cursors.AppStarting;
                    }
                }
                else
                {
                    this.btnExecute.Enabled = false;

                    this.fileScanner.AbortScanning();
                    this.fileScanner.ScanningStarted -= this.OnScanningStarted;
                    this.fileScanner.FileScanCompleted -= this.OnFileScanCompleted;
                    this.fileScanner.ScanningCompleted -= this.OnScanningCompleted;
                    this.fileScanner = null;

                    this.SwapButtonLabel(this.btnExecute);
                    this.Cursor = Cursors.Default;
                    this.btnExecute.Enabled = true;
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);

                if (this.fileScanner != null)
                {
                    this.fileScanner.AbortScanning();
                    this.fileScanner.ScanningStarted -= this.OnScanningStarted;
                    this.fileScanner.FileScanCompleted -= this.OnFileScanCompleted;
                    this.fileScanner.ScanningCompleted -= this.OnScanningCompleted;
                    this.fileScanner = null;
                }

                this.sblFileValue.Text = "Failed!";
                this.btnExecute.Enabled = true;
                this.Cursor = Cursors.Default;
            }
        }

        #endregion // Main form control event handlers.

        #region Toolbar button click handler.

        private void OnExitButtonClick(object sender, EventArgs args)
        {
            if (this.HandleDirty())
            {
                this.IsDirty = false;
                Application.Exit();
            }
        }

        private void OnNewButtonClick(object sender, EventArgs args)
        {
            if (this.HandleDirty())
            {
                this.NewEnvironment();
            }
        }

        private void OnOpenButtonClick(object sender, EventArgs args)
        {
            if (this.HandleDirty())
            {
                OpenFileDialog dialog = new OpenFileDialog();
                dialog.AddExtension = true;
                dialog.CheckFileExists = true;
                dialog.CheckPathExists = true;
                dialog.DefaultExt = MainForm.OPEN_SAVE_EXTENSION;
                dialog.Multiselect = false;
                dialog.RestoreDirectory = true;
                dialog.ShowHelp = false;
                dialog.ShowReadOnly = false;
                dialog.ValidateNames = true;
                dialog.Filter = MainForm.OPEN_SAVE_FILTER;
                dialog.InitialDirectory = this.GetValidDirectory(Settings.Instance.General.LastEnvironment);
                if (DialogResult.OK == dialog.ShowDialog())
                {
                    this.OpenEnvironment(dialog.FileName);
                }
            }
        }

        private void OnSaveButtonClick(object sender, EventArgs args)
        {
            if (this.IsDirty)
            {
                SaveFileDialog dialog = new SaveFileDialog();
                dialog.AddExtension = true;
                dialog.CheckFileExists = false;
                dialog.CheckPathExists = true;
                dialog.DefaultExt = MainForm.OPEN_SAVE_EXTENSION;
                dialog.RestoreDirectory = true;
                dialog.ShowHelp = false;
                dialog.ValidateNames = false;
                dialog.Filter = MainForm.OPEN_SAVE_FILTER;
                dialog.InitialDirectory = this.GetValidDirectory(Settings.Instance.General.LastEnvironment);
                dialog.FileName = this.GetValidFilename(Settings.Instance.General.LastEnvironment);
                if (DialogResult.OK == dialog.ShowDialog())
                {
                    this.SaveEnvironment(dialog.FileName);
                }
            }
        }

        private void OnIncludeButtonClick(object sender, EventArgs args)
        {
            this.tbbInclude.ShowDropDown();
        }

        private void OnIncludeButtonMenuOpening(object sender, EventArgs args)
        {
            bool validPath = this.IsValidBasePath;
            this.tbmIncludeSelectFolders.Enabled = validPath;
            this.tbmIncludeClearFolders.Enabled = this.Environment.IsSelectedFolders;
            this.tbmIncludeSelectFiles.Enabled = validPath;
            this.tbmIncludeClearFiles.Enabled = this.Environment.IsSelectedFiles;
            this.tbmIncludeSearchFolders.Enabled = this.Environment.IsSelectedFolders && this.Environment.IsSelectedFiles;
        }

        private void OnIncludeSelectFoldersMenuClick(object sender, EventArgs args)
        {
            string baseBath;
            if (this.IsBasePathValidated(out baseBath))
            {
                FolderSelectorDialog dialog = new FolderSelectorDialog(baseBath, this.Environment.SelectedFolders);
                if (DialogResult.OK == dialog.ShowDialog())
                {
                    this.Environment.SelectedFolders = dialog.SelectedItems;
                    this.IsDirty = true;
                }
            }
        }

        private void OnIncludeClearFoldersMenuClick(object sender, EventArgs args)
        {
            this.Environment.SelectedFolders = null;
        }

        private void OnIncludeSelectFilesMenuClick(object sender, EventArgs args)
        {
            string baseBath;
            if (this.IsBasePathValidated(out baseBath) && this.cmbFileGroups.SelectedItem != null)
            {
                string[] pattern = ((FileFilterGroup)this.cmbFileGroups.SelectedItem).SearchPattern.Values;
                FileSelectorDialog dialog = new FileSelectorDialog(baseBath, pattern, this.Environment.SelectedFiles);
                if (DialogResult.OK == dialog.ShowDialog())
                {
                    this.Environment.SelectedFiles = dialog.SelectedItems;
                    this.IsDirty = true;
                }
            }
        }

        private void OnIncludeClearFilesMenuClick(object sender, EventArgs args)
        {
            this.Environment.SelectedFiles = null;
        }

        private void OnIncludeSearchFoldersMenuClick(object sender, EventArgs args)
        {
            this.IsDirty = true;
        }

        private void OnFilterButtonClick(object sender, EventArgs args)
        {
            FileFilterGroup lastGroup = this.cmbFileGroups.SelectedItem as FileFilterGroup;
            FileFilterGroupDialog dialog = new FileFilterGroupDialog(this.Environment.GroupCollection);
            if (DialogResult.OK == dialog.ShowDialog())
            {
                this.Environment.GroupCollection = dialog.Groups;
                this.cmbFileGroups.DataSource = this.Environment.GroupCollection;
                this.SetSelectedFileGroup(lastGroup);
                this.IsDirty = true;
            }
        }

        private void OnSettingsButtonClick(object sender, EventArgs args)
        {
            // RFU
        }

        private void OnAboutButtonClick(object sender, EventArgs args)
        {
            this.tbbAbout.ShowDropDown();
        }

        private void OnAboutHelpMenuClick(object sender, EventArgs args)
        {
            Program.ShowHelp();
        }

        private void OnAboutAboutMenuClick(object sender, EventArgs args)
        {
            AboutBox about = new AboutBox();
            about.ShowDialog();
        }

        private void OnReportButtonClick(object sender, EventArgs args)
        {
            this.tbbReport.ShowDropDown();
        }

        private void OnReportBriefHtmlMenuClick(object sender, EventArgs args)
        {
            SaveFileDialog dialog = new SaveFileDialog();
            dialog.AddExtension = true;
            dialog.CheckFileExists = false;
            dialog.CheckPathExists = true;
            dialog.DefaultExt = MainForm.REPORT_SAVE_EXTENSION;
            dialog.RestoreDirectory = true;
            dialog.ShowHelp = false;
            dialog.ValidateNames = false;
            dialog.Filter = MainForm.REPORT_SAVE_FILTER;
            dialog.InitialDirectory = this.GetValidDirectory(Settings.Instance.General.LastReportFile);
            dialog.FileName = this.GetValidFilename(Settings.Instance.General.LastReportFile, MainForm.REPORT_SAVE_EXTENSION);
            if (DialogResult.OK == dialog.ShowDialog())
            {
                this.SaveReport(dialog.FileName, ReportGenerator.ReportType.Brief);
            }
        }

        private void OnReportFullHtmlMenuClick(object sender, EventArgs args)
        {
            SaveFileDialog dialog = new SaveFileDialog();
            dialog.AddExtension = true;
            dialog.CheckFileExists = false;
            dialog.CheckPathExists = true;
            dialog.DefaultExt = MainForm.REPORT_SAVE_EXTENSION;
            dialog.RestoreDirectory = true;
            dialog.ShowHelp = false;
            dialog.ValidateNames = false;
            dialog.Filter = MainForm.REPORT_SAVE_FILTER;
            dialog.InitialDirectory = this.GetValidDirectory(Settings.Instance.General.LastReportFile);
            dialog.FileName = this.GetValidFilename(Settings.Instance.General.LastReportFile, MainForm.REPORT_SAVE_EXTENSION);
            if (DialogResult.OK == dialog.ShowDialog())
            {
                this.SaveReport(dialog.FileName, ReportGenerator.ReportType.Full);
            }
        }

        #endregion // Toolbar button click handler.

        #region File Scanner event handlers.

        private delegate void OnScanningStartedDelegate(object sender, ScanningStartedArgs args);
        private void OnScanningStarted(object sender, ScanningStartedArgs args)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new OnScanningStartedDelegate(OnScanningStarted), new object[] { sender, args });
            }
            else
            {
                this.btnExecute.Enabled = true;
                this.prbExecution.Total = args.Count;
                this.prbExecution.Value = 0;
                this.sblScannedValue.Text = this.prbExecution.Value.ToString("N0", NumberFormatInfo.CurrentInfo);
            }
        }

        private delegate void OnFileScanCompletedDelegate(object sender, FileScanCompletedArgs args);
        private void OnFileScanCompleted(object sender, FileScanCompletedArgs args)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new OnFileScanCompletedDelegate(OnFileScanCompleted), new object[] { sender, args });
            }
            else
            {
                this.prbExecution.Increment();
                this.sblScannedValue.Text = this.prbExecution.Value.ToString("N0", NumberFormatInfo.CurrentInfo);

                try { this.sblFileValue.Text = Path.GetFileName(args.Filename); }
                catch (Exception exception) { Debug.WriteLine(exception); }
            }
        }

        private delegate void OnScanningCompletedDelegate(object sender, EventArgs args);
        private void OnScanningCompleted(object sender, EventArgs args)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new OnScanningCompletedDelegate(OnScanningCompleted), new object[] { sender, args });
            }
            else
            {
                this.fileScanner.ScanningStarted -= this.OnScanningStarted;
                this.fileScanner.FileScanCompleted -= this.OnFileScanCompleted;
                this.fileScanner.ScanningCompleted -= this.OnScanningCompleted;
                this.scanningResults = new ScanningResults(this.fileScanner);
                this.fileScanner = null;

                this.sblFileValue.Text = "Done!";

                this.SwapButtonLabel(this.btnExecute);
                this.Cursor = Cursors.Default;

                this.ShowResults();
                this.UpdateControls();
            }
        }

        #endregion // File Scanner event handlers.

        #region Main form event handlers.

        private void OnMainFormLoad(object sender, EventArgs args)
        {
            // If loading configuration fails the form starts with default size and centred!
            if (Settings.Instance.Load())
            {
                Settings settings = Settings.Instance;

                // Set size before location because location depends on size!
                this.WindowState = settings.MainWindow.State;
                this.Size = settings.MainWindow.Size;
                this.Location = settings.MainWindow.Location;

                // Bad but needed to adjust visible controls before 
                // loading a file (which takes some time).
                Application.DoEvents();
            }
        }

        private void OnMainFormClosing(object sender, FormClosingEventArgs args)
        {
            try
            {
                if (this.fileScanner != null)
                {
                    this.fileScanner.AbortScanning();
                    this.fileScanner.ScanningStarted -= this.OnScanningStarted;
                    this.fileScanner.FileScanCompleted -= this.OnFileScanCompleted;
                    this.fileScanner.ScanningCompleted -= this.OnScanningCompleted;
                    this.fileScanner = null;
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }

#if !DEBUG
            args.Cancel = !this.HandleDirty();
#endif
        }

        private void OnMainFormClosed(object sender, FormClosedEventArgs args)
        {
            // Set size before location because location depends on size!
            Settings.Instance.MainWindow.State = this.WindowState;
            Settings.Instance.MainWindow.Size = this.Size;
            Settings.Instance.MainWindow.Location = this.Location;

            Settings.Instance.Save();
        }

        #endregion // Main form event handlers.
    }
}
