/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.Globalization;
using System.ComponentModel;
using System.Drawing.Drawing2D;

namespace plexdata.LineCounter
{
    public class ProgressBar : Panel
    {
        private long value = 0;
        private long steps = 1;
        private long total = 100;

        private bool textEnabled = true;
        private bool nullTextEnabled = true;
        private bool vertical = false;
        private bool textVertical = false;
        private bool percentage = true;

        private StringFormat horzFormat;
        private StringFormat vertFormat;

        public ProgressBar()
            : base()
        {
            this.DoubleBuffered = true;

            this.horzFormat = new StringFormat();
            this.horzFormat.Alignment = StringAlignment.Center;
            this.horzFormat.LineAlignment = StringAlignment.Center;

            this.vertFormat = new StringFormat(StringFormatFlags.DirectionVertical);
            this.vertFormat.Alignment = StringAlignment.Center;
            this.vertFormat.LineAlignment = StringAlignment.Center;

            this.Font = new Font("Microsoft Sans Serif", 8.25F, FontStyle.Bold);
            this.ForeColor = Color.CornflowerBlue;
            this.BackColor = Color.White;
            this.BorderStyle = BorderStyle.Fixed3D;
        }

        [DefaultValue(0)]
        [Category("Progress")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public long Value
        {
            get { return this.value; }
            set
            {
                this.value = (value <= this.total) ? value : this.total;
                this.Invalidate();
            }
        }

        [DefaultValue(1)]
        [Category("Progress")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public long Steps
        {
            get { return this.steps; }
            set
            {
                this.steps = (value <= this.total) ? value : this.total;
                this.Invalidate();
            }
        }

        [DefaultValue(100)]
        [Category("Progress")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public long Total
        {
            get { return this.total; }
            set { this.total = value; }
        }

        [DefaultValue(true)]
        [Category("Progress")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public bool TextEnabled
        {
            get { return this.textEnabled; }
            set
            {
                if (this.textEnabled != value)
                {
                    this.textEnabled = value;
                    this.Invalidate();
                }
            }
        }

        [DefaultValue(true)]
        [Category("Progress")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public bool NullTextEnabled
        {
            get { return this.nullTextEnabled; }
            set
            {
                if (this.nullTextEnabled != value)
                {
                    this.nullTextEnabled = value;
                    this.Invalidate();
                }
            }
        }

        [DefaultValue(false)]
        [Category("Progress")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public bool Vertical
        {
            get { return this.vertical; }
            set
            {
                if (this.vertical != value)
                {
                    this.vertical = value;
                    this.Invalidate();
                }
            }
        }

        [DefaultValue(false)]
        [Category("Progress")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public bool TextVertical
        {
            get { return this.textVertical; }
            set
            {
                if (this.textVertical != value)
                {
                    this.textVertical = value;
                    this.Invalidate();
                }
            }
        }

        [DefaultValue(true)]
        [Category("Progress")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public bool Percentage
        {
            get { return this.percentage; }
            set
            {
                if (this.percentage != value)
                {
                    this.percentage = value;
                    this.Invalidate();
                }
            }
        }

        [DefaultValue(BorderStyle.Fixed3D)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public new BorderStyle BorderStyle { get; set; }

        [DefaultValue(typeof(Color), "CornflowerBlue")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public new Color ForeColor { get; set; }

        [DefaultValue(typeof(Color), "White")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public new Color BackColor { get; set; }

        [DefaultValue(typeof(Font), "Microsoft Sans Serif, 8.25pt, style=Bold")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public new Font Font { get; set; }

        protected override CreateParams CreateParams
        {
            get
            {
                // Very smart because client rectangle is also adjusted! 
                // See also: http://support.microsoft.com/kb/316574
                const int WS_BORDER = unchecked((int)0x00800000);
                const int WS_EX_STATICEDGE = unchecked((int)0x00020000);

                CreateParams createParams = base.CreateParams;
                createParams.ExStyle &= (~WS_EX_STATICEDGE);
                createParams.Style &= (~WS_BORDER);

                switch (this.BorderStyle)
                {
                    case BorderStyle.Fixed3D:
                        createParams.ExStyle |= WS_EX_STATICEDGE;
                        break;
                    case BorderStyle.FixedSingle:
                        createParams.Style |= WS_BORDER;
                        break;
                }
                return createParams;
            }
        }

        public void Increment()
        {
            this.Increment(this.Steps);
        }

        public void Increment(long steps)
        {
            this.Value += steps;
        }

        protected override void OnPaintBackground(PaintEventArgs args)
        {
            base.OnPaintBackground(args);
            using (Brush brush = new SolidBrush(this.BackColor))
            {
                args.Graphics.FillRectangle(brush, this.ClientRectangle);
            }
        }

        protected override void OnPaint(PaintEventArgs args)
        {
            base.OnPaint(args);

            Rectangle filled = new Rectangle(this.ClientRectangle.Location, this.ClientRectangle.Size);
            Rectangle remain = new Rectangle(filled.Location, filled.Size);

            decimal divisor = (decimal)(this.total != 0 ? this.total : 1);
            if (this.vertical)
            {
                decimal height = (decimal)filled.Height * ((decimal)this.value) / divisor;
                if (height <= filled.Height) { filled.Y = filled.Height - (int)height; }

                remain.Height -= (int)height;
            }
            else
            {
                decimal width = (decimal)filled.Width * ((decimal)this.value) / divisor;
                if (width <= filled.Width) { filled.Width = (int)width; }

                remain.X = filled.Width;
            }

            // Draw progress rectangle. 
            using (Brush brush = new SolidBrush(this.ForeColor))
            {
                args.Graphics.FillRectangle(brush, filled);
            }

            // Draw text.
            if (this.textEnabled)
            {
                if (!this.nullTextEnabled && this.value == 0) { return; }

                string label = String.Empty;
                if (this.percentage)
                {
                    long percent = (long)((this.value * 100) / divisor);
                    label = percent.ToString() + "%";
                }
                else
                {
                    label = this.value.ToString("N0", NumberFormatInfo.CurrentInfo);
                }

                StringFormat format = this.horzFormat;
                if (this.textVertical) { format = this.vertFormat; }

                using (Brush brush = new SolidBrush(this.BackColor))
                {
                    Region region = new Region(this.ClientRectangle);
                    region.Intersect(filled);
                    args.Graphics.Clip = region;
                    args.Graphics.DrawString(label, this.Font, brush, this.ClientRectangle, format);
                }

                using (Brush brush = new SolidBrush(this.ForeColor))
                {
                    Region region = new Region(this.ClientRectangle);
                    region.Intersect(remain);
                    args.Graphics.Clip = region;
                    args.Graphics.DrawString(label, this.Font, brush, this.ClientRectangle, format);
                }
            }
        }

        protected override void OnSizeChanged(EventArgs args)
        {
            base.OnSizeChanged(args);
            this.Invalidate();
        }

        protected override void OnFontChanged(EventArgs args)
        {
            base.OnFontChanged(args);
            this.Invalidate();
        }

        protected override void OnBackColorChanged(EventArgs args)
        {
            base.OnBackColorChanged(args);
            this.Invalidate();
        }

        protected override void OnForeColorChanged(EventArgs args)
        {
            base.OnForeColorChanged(args);
            this.Invalidate();
        }
    }
}
