﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Diagnostics;
using System.Globalization;

namespace plexdata.LineCounter
{
    public static class ReportGenerator
    {
        public enum ReportType
        {
            Brief,
            Full
        }

        public static string GetDisplayReport(ScanningResult scanningResult)
        {
            try
            {
                if (scanningResult != null)
                {
                    return String.Format(HTML_BODY,
                        CreateMetricTable(scanningResult) +
                        CreateRatioTable(scanningResult) +
                        CreateSchemeTable(scanningResult) +
                        CreateErrorTable(scanningResult));
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            return String.Empty;
        }

        public static string GetExportReport(ScanningResults scanningResults, ReportType reportType)
        {
            try
            {
                if (scanningResults != null)
                {
                    if (reportType == ReportType.Brief)
                    {
                        // Header and summary.
                        string htmlBody =
                            String.Format(TITLE_HEADER, "Brief Report") +
                            CreateExportHeader(scanningResults, ReportType.Brief) +
                            CreateMetricTable(scanningResults) +
                            CreateRatioTable(scanningResults) +
                            CreateSchemeTable(scanningResults);

                        // File type dependent reports.
                        foreach (ScanningResult scanningResult in scanningResults.TypeResults)
                        {
                            htmlBody +=
                                CreateMetricTable(scanningResult) +
                                CreateRatioTable(scanningResult) +
                                CreateSchemeTable(scanningResult);
                        }

                        return String.Format(HTML_BODY, htmlBody);
                    }
                    else if (reportType == ReportType.Full)
                    {
                        // Header and summary.
                        string htmlBody =
                            String.Format(TITLE_HEADER, "Full Report") +
                            CreateExportHeader(scanningResults, ReportType.Full) +
                            CreateMetricTable(scanningResults) +
                            CreateRatioTable(scanningResults) +
                            CreateSchemeTable(scanningResults);

                        // File type dependent reports.
                        foreach (ScanningResult scanningResult in scanningResults.TypeResults)
                        {
                            htmlBody +=
                                CreateMetricTable(scanningResult) +
                                CreateRatioTable(scanningResult) +
                                CreateSchemeTable(scanningResult) +
                                CreateFileItems(scanningResult);
                        }

                        return String.Format(HTML_BODY, htmlBody);
                    }
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
            return String.Empty;
        }

        private static string CreateExportHeader(ScanningResult scanningResult, ReportType reportType)
        {
            string bodyValue = String.Empty;
            string tableValues = String.Empty;

            #region Export tabel settings.

            tableValues += String.Format(EXPORT_TABLE_LINE, "Program:", AboutBox.Title);

            tableValues += String.Format(EXPORT_TABLE_LINE, "Version:", AboutBox.Version);

            tableValues += String.Format(EXPORT_TABLE_LINE, "Copyright:", ReplaceHTMLCharacter(AboutBox.Copyright));

            tableValues += String.Format(EXPORT_TABLE_LINE, "Generated:", FormatDateTime(DateTime.Now));

            tableValues += String.Format(EXPORT_TABLE_LINE, "Report:", reportType.ToString());

            if (scanningResult is ScanningResults)
            {
                tableValues += String.Format(
                    EXPORT_TABLE_LINE, "Filter" + FIXED_SPACE + "Group:",
                    (scanningResult as ScanningResults).FileScanner.Filter.Name);

                tableValues += String.Format(
                    EXPORT_TABLE_LINE, "Description:",
                    (scanningResult as ScanningResults).FileScanner.Filter.Description);

                tableValues += String.Format(
                    EXPORT_TABLE_LINE, "File" + FIXED_SPACE + "Pattern:",
                    (scanningResult as ScanningResults).FileScanner.Filter.SearchPattern);
            }

            bodyValue =
                String.Format(EXPORT_TABLE_BODY, tableValues);

            #endregion // Export tabel settings.

            return bodyValue;
        }

        private static string CreateMetricTable(ScanningResult scanningResult)
        {
            string bodyValue = String.Empty;
            string headerValue = String.Empty;
            string tableValues = String.Empty;

            #region Metric tabel settings.

            if (scanningResult is ScanningResults)
            {
                headerValue = "Overall Summary";
            }
            else
            {
                headerValue = String.Format("Overview of {0} Files", scanningResult.Label);
            }

            tableValues += String.Format(
                METRIC_TABLE_LINE, "Total Lines", FormatNumber(scanningResult.TotalLines));

            tableValues += String.Format(
                METRIC_TABLE_LINE, "Number of Files", GetFilesCount(scanningResult));

            tableValues += String.Format(
                METRIC_TABLE_LINE, "Lines with Code", FormatNumber(scanningResult.LinesWithCode));

            tableValues += String.Format(
                METRIC_TABLE_LINE, String.Format(FOOTNOTE_ITEM, "Comment Lines", "1"),
                FormatNumber(scanningResult.CommentLines));

            tableValues += String.Format(
                METRIC_TABLE_LINE, "Empty Lines", FormatNumber(scanningResult.EmptyLines));

            if (scanningResult.StringLiterals > 0)
            {
                tableValues += String.Format(
                    METRIC_TABLE_LINE, "String Literals", FormatNumber(scanningResult.StringLiterals));
            }

            if (scanningResult.CharacterLiterals > 0)
            {
                tableValues += String.Format(
                    METRIC_TABLE_LINE, "Character Literals", FormatNumber(scanningResult.CharacterLiterals));
            }

            tableValues += String.Format(
                METRIC_TABLE_LINE, "Scan Duration", GetScanDuration(scanningResult.ScanDuration));

            bodyValue =
                String.Format(MAIN_HEADER, headerValue) +
                String.Format(METRIC_TABLE_BODY, "Metric", "Value", tableValues) +
                String.Format(FOOTNOTE_TEXT, "1", "Lines that only contain comments.") +
                LINE_BREAK;

            #endregion // Metric tabel settings.

            return bodyValue;
        }

        private static string CreateRatioTable(ScanningResult scanningResult)
        {
            string bodyValue = String.Empty;
            string tableValues = String.Empty;

            #region Ratio tabel settings.

            tableValues += String.Format(
                RATIO_TABLE_LINE, "<i>Lines with Code</i> per <i>Total Lines</i>",
                FormatRatio(scanningResult.LinesWithCode, scanningResult.TotalLines));

            tableValues += String.Format(
                RATIO_TABLE_LINE, "<i>Lines with Code</i> per <i>Comment Lines</i>",
                FormatRatio(scanningResult.LinesWithCode, scanningResult.CommentLines));

            tableValues += String.Format(
                RATIO_TABLE_LINE, "<i>Lines with Code</i> per <i>Empty Lines</i>",
                FormatRatio(scanningResult.LinesWithCode, scanningResult.EmptyLines));

            tableValues += String.Format(
                RATIO_TABLE_LINE, "<i>Lines with Code</i> per <i>(Empty and Comment Lines)</i>",
                FormatRatio(scanningResult.LinesWithCode, scanningResult.EmptyLines + scanningResult.CommentLines));

            bodyValue =
                String.Format(RATIO_TABLE_BODY, "Comparison", "Percentage", tableValues) +
                ((scanningResult is ScanningResults) ? String.Empty : LINE_BREAK);

            #endregion // Ratio tabel settings.

            return bodyValue;
        }

        private static string CreateSchemeTable(ScanningResult scanningResult)
        {
            string bodyValue = String.Empty;
            string tableValues = String.Empty;

            #region Scheme tabel settings.

            if (scanningResult.FindingResults.Length > 0)
            {
                foreach (FindingResult findings in scanningResult.FindingResults)
                {
                    tableValues += String.Format(
                        SCHEME_TABLE_LINE, ReplaceHTMLCharacter(findings.CommentScheme),
                        FormatNumber(findings.PureCommentCount),
                        FormatNumber(findings.PureCommentLines),
                        FormatNumber(findings.CodeCommentCount),
                        FormatNumber(findings.CodeCommentLines));
                }

                bodyValue =
                    String.Format(SCHEME_TABLE_BODY,
                        "Comment Schemes",
                        "Pure Comments",
                        String.Format(FOOTNOTE_ITEM, "Hits", "1"),
                        String.Format(FOOTNOTE_ITEM, "Lines", "2"),
                        "Code Comments",
                        String.Format(FOOTNOTE_ITEM, "Hits", "1"),
                        String.Format(FOOTNOTE_ITEM, "Lines", "2"),
                        tableValues) +
                        String.Format(FOOTNOTE_TEXT, "1", "Number of occurrences of this comment scheme.") +
                        String.Format(FOOTNOTE_TEXT, "2", "Number of countable lines per comment scheme.");
            }

            #endregion // Scheme tabel settings.

            return bodyValue;
        }

        private static string CreateErrorTable(ScanningResult scanningResult)
        {
            string bodyValue = String.Empty;
            string headerValue = String.Empty;
            string tableValues = String.Empty;

            #region Error tabel settings.

            if (scanningResult is ScanningResults)
            {
                // Summary table does not have an 
                // error table! Nothing to do.
                return String.Empty;
            }

            if (scanningResult.ContentScanners.Length > 0)
            {
                foreach (ContentScanner contentScanner in scanningResult.ContentScanners)
                {
                    if (contentScanner.Exception != null)
                    {
                        tableValues += String.Format(
                            ERROR_TABLE_LINE,
                            Path.GetFileName(contentScanner.Filename),
                            contentScanner.Exception.Message,
                            Path.GetDirectoryName(contentScanner.Filename));
                    }
                }

                if (!String.IsNullOrEmpty(tableValues))
                {
                    bodyValue =
                        String.Format(ERROR_HEADER, String.Format("Error List")) +
                        String.Format(ERROR_TABLE_BODY, tableValues);
                }
            }

            #endregion // Error tabel settings.

            return bodyValue;
        }

        private static string CreateFileItems(ScanningResult scanningResult)
        {
            string bodyValue = String.Empty;
            string headerValue = String.Empty;
            string tableValues = String.Empty;
            string infoValues = String.Empty;
            string metricValues = String.Empty;
            string ratioValues = String.Empty;
            string schemeValues = String.Empty;

            if (scanningResult.ContentScanners.Length > 0)
            {
                foreach (ContentScanner contentScanner in scanningResult.ContentScanners)
                {
                    headerValue = String.Format(FILE_HEADER,
                        String.Format("Overview of File {0}", Path.GetFileName(contentScanner.Filename)));

                    #region Info tabel settings.

                    tableValues = String.Format(
                        INFO_TABLE_LINE, "Location:", Path.GetDirectoryName(contentScanner.Filename));

                    FileInfo fileInfo = null;
                    try
                    {
                        fileInfo = new FileInfo(contentScanner.Filename);
                    }
                    catch (Exception exception)
                    {
                        Debug.WriteLine(exception);
                    }

                    if (fileInfo != null)
                    {
                        tableValues += String.Format(
                            INFO_TABLE_LINE, "Created:", FormatDateTime(fileInfo.CreationTimeUtc));

                        tableValues += String.Format(
                            INFO_TABLE_LINE, "Changed:", FormatDateTime(fileInfo.LastWriteTimeUtc));

                        tableValues += String.Format(
                            INFO_TABLE_LINE, "Size:", FormatFileSize(fileInfo.Length));

                        tableValues += String.Format(
                            INFO_TABLE_LINE, "Attributes:", fileInfo.Attributes.ToString());
                    }

                    tableValues += String.Format(
                        INFO_TABLE_LINE, "Linefeed:", contentScanner.Linefeed);

                    tableValues += String.Format(
                        INFO_TABLE_LINE, "Encoding:",
                        (contentScanner.Encoding != null ? contentScanner.Encoding.EncodingName : "Not detected!"));

                    if (contentScanner.Exception != null)
                    {
                        tableValues += String.Format(
                            INFO_TABLE_LINE, "Error:", contentScanner.Exception.Message);
                    }

                    infoValues = String.Format(INFO_TABLE_BODY, tableValues) + LINE_BREAK;

                    #endregion // Info tabel settings.

                    if (contentScanner.Exception == null)
                    {
                        #region Metric tabel settings.

                        tableValues = String.Format(
                            METRIC_TABLE_LINE, "Total Lines", FormatNumber(contentScanner.TotalLines));

                        tableValues += String.Format(
                            METRIC_TABLE_LINE, "Lines with Code", FormatNumber(contentScanner.LinesWithCode));

                        tableValues += String.Format(
                            METRIC_TABLE_LINE, String.Format(FOOTNOTE_ITEM, "Comment Lines", "1"),
                            FormatNumber(contentScanner.CommentLines));

                        tableValues += String.Format(
                            METRIC_TABLE_LINE, "Empty Lines", FormatNumber(contentScanner.EmptyLines));

                        if (contentScanner.StringLiterals > 0)
                        {
                            tableValues += String.Format(
                                METRIC_TABLE_LINE, "String Literals", FormatNumber(contentScanner.StringLiterals));
                        }

                        if (contentScanner.CharacterLiterals > 0)
                        {
                            tableValues += String.Format(
                                METRIC_TABLE_LINE, "Character Literals", FormatNumber(contentScanner.CharacterLiterals));
                        }

                        tableValues += String.Format(
                            METRIC_TABLE_LINE, "Scan Duration", GetScanDuration(contentScanner.ScanDuration));

                        metricValues =
                            String.Format(METRIC_TABLE_BODY, "Metric", "Value", tableValues) +
                            String.Format(FOOTNOTE_TEXT, "1", "Lines that only contain comments.") +
                            LINE_BREAK;

                        #endregion // Metric tabel settings.

                        #region Ratio tabel settings.

                        tableValues = String.Format(
                            RATIO_TABLE_LINE, "<i>Lines with Code</i> per <i>Total Lines</i>",
                            FormatRatio(contentScanner.LinesWithCode, contentScanner.TotalLines));

                        tableValues += String.Format(
                            RATIO_TABLE_LINE, "<i>Lines with Code</i> per <i>Comment Lines</i>",
                            FormatRatio(contentScanner.LinesWithCode, contentScanner.CommentLines));

                        tableValues += String.Format(
                            RATIO_TABLE_LINE, "<i>Lines with Code</i> per <i>Empty Lines</i>",
                            FormatRatio(contentScanner.LinesWithCode, contentScanner.EmptyLines));

                        tableValues += String.Format(
                            RATIO_TABLE_LINE, "<i>Lines with Code</i> per <i>(Empty and Comment Lines)</i>",
                            FormatRatio(contentScanner.LinesWithCode, contentScanner.EmptyLines + contentScanner.CommentLines));

                        ratioValues =
                            String.Format(RATIO_TABLE_BODY, "Comparison", "Percentage", tableValues) +
                            LINE_BREAK;

                        #endregion // Ratio tabel settings.

                        #region Scheme tabel settings.

                        if (contentScanner.Findings.Count > 0)
                        {
                            tableValues = String.Empty;

                            foreach (CommentFindings findings in contentScanner.Findings)
                            {
                                tableValues += String.Format(
                                    SCHEME_TABLE_LINE,
                                    ReplaceHTMLCharacter(FindingResult.BuildCommentScheme(findings.Scheme)),
                                    FormatNumber(findings.PureCommentCount),
                                    FormatNumber(findings.PureCommentLines),
                                    FormatNumber(findings.CodeCommentCount),
                                    FormatNumber(findings.CodeCommentLines));
                            }

                            schemeValues =
                                String.Format(SCHEME_TABLE_BODY,
                                    "Comment Schemes",
                                    "Pure Comments",
                                    String.Format(FOOTNOTE_ITEM, "Hits", "1"),
                                    String.Format(FOOTNOTE_ITEM, "Lines", "2"),
                                    "Code Comments",
                                    String.Format(FOOTNOTE_ITEM, "Hits", "1"),
                                    String.Format(FOOTNOTE_ITEM, "Lines", "2"),
                                    tableValues) +
                                    String.Format(FOOTNOTE_TEXT, "1", "Number of occurrences of this comment scheme.") +
                                    String.Format(FOOTNOTE_TEXT, "2", "Number of countable lines per comment scheme.");
                        }

                        #endregion // Scheme tabel settings.
                    }

                    bodyValue +=
                        headerValue + infoValues + metricValues + ratioValues + schemeValues;
                }
            }

            return bodyValue;
        }

        private static string GetScanDuration(TimeSpan duration)
        {
            double milliseconds = duration.TotalMilliseconds;
            if (milliseconds < 1)
            {
                return "< 1 ms";
            }
            else
            {
                return milliseconds.ToString("N0", NumberFormatInfo.CurrentInfo) + " ms";
            }
        }

        private static string GetFilesCount(ScanningResult scanningResult)
        {
            int count = 0;
            if (scanningResult is ScanningResults)
            {
                foreach (ScanningResult result in (scanningResult as ScanningResults).TypeResults)
                {
                    count += result.FilesCount;
                }
            }
            else
            {
                count = scanningResult.FilesCount;
            }
            return count.ToString("N0", NumberFormatInfo.CurrentInfo);
        }

        private static string FormatNumber(long value)
        {
            return value.ToString("N0", NumberFormatInfo.CurrentInfo);
        }

        private static string FormatRatio(long value, long comparison)
        {
            // Avoid division by zero!
            if (value == 0 || comparison == 0)
            {
                return ((decimal)0).ToString("P2", NumberFormatInfo.CurrentInfo);
            }
            else
            {
                return ((decimal)value / (decimal)comparison).ToString("P2", NumberFormatInfo.CurrentInfo);
            }
        }

        private static string ReplaceHTMLCharacter(string value)
        {
            string[,] characters = new string[,] { 
                { "\"", "&quot;" }, { "&", "&amp;" }, { "<", "&lt;" }, { ">", "&gt;" }, { "©", "&copy;" }, };

            for (int index = 0; index <= characters.GetUpperBound(0); index++)
            {
                value = value.Replace(characters[index, 0], characters[index, 1]);
            }

            return value;
        }

        private static string FormatDateTime(DateTime value, bool utc = true)
        {
            if (utc) { value = value.ToUniversalTime(); }

            return String.Format(
                "{0:D4}-{1:D2}-{2:D2} {3:D2}:{4:D2}:{5:D2}{6}",
                value.Year, value.Month, value.Day,
                value.Hour, value.Minute, value.Second,
                (utc ? " (UTC)" : String.Empty));
        }

        private static string FormatFileSize(long value)
        {
            const long KiB = 1024;
            const long MiB = 1048576;
            const long GiB = 1073741824;
            const long TiB = 1099511627776;
            const long EiB = 1125899906842624;

            if (value >= EiB)
            {
                return (value / EiB).ToString("N0", NumberFormatInfo.CurrentInfo) + " EiB";
            }
            else if (value >= TiB)
            {
                return (value / TiB).ToString("N0", NumberFormatInfo.CurrentInfo) + " TiB";
            }
            else if (value >= GiB)
            {
                return (value / GiB).ToString("N0", NumberFormatInfo.CurrentInfo) + " GiB";
            }
            else if (value >= MiB)
            {
                return (value / MiB).ToString("N0", NumberFormatInfo.CurrentInfo) + " MiB";
            }
            else if (value >= KiB)
            {
                return (value / KiB).ToString("N0", NumberFormatInfo.CurrentInfo) + " KiB";
            }
            else
            {
                return value.ToString("N0", NumberFormatInfo.CurrentInfo) + " Bytes";
            }
        }

        #region HTML tag definition section.

        private static string HTML_BODY =
            "<!DOCTYPE HTML PUBLIC \" -//W3C//DTD HTML 4.0 Transitional//EN\">" +
            "<html>" +
                "<head>" +
                    "<style type=\"text/css\">" +
                        ".error_table_line" +
                        "{{" +
                            "font-size:0.8em;" +
                        "}}" +
                        ".info_table_line" +
                        "{{" +
                            "font-size:0.8em;" +
                        "}}" +
                        ".export_table_line" +
                        "{{" +
                            "font-size:0.8em;" +
                        "}}" +
                        ".metric_table_header" +
                        "{{" +
                            "font-size:0.8em; background-color:#A0BAD7;" +
                        "}}" +
                        ".metric_table_line" +
                        "{{" +
                            "font-size:0.8em; border-bottom-style:solid; border-bottom-width:1px; border-bottom-color:#C0C0C0;" +
                        "}}" +
                        ".ratio_table_header" +
                        "{{" +
                            "font-size:0.8em; background-color:#FF6D72;" +
                        "}}" +
                        ".ratio_table_line" +
                        "{{" +
                            "font-size:0.8em; border-bottom-style:solid; border-bottom-width:1px; border-bottom-color:#C0C0C0;" +
                        "}}" +
                        ".scheme_table_header" +
                        "{{" +
                            "font-size:0.8em; background-color:#FFF575;" +
                        "}}" +
                        ".scheme_table_line" +
                        "{{" +
                            "font-size:0.8em; border-bottom-style:solid; border-bottom-width:1px; border-bottom-color:#C0C0C0;" +
                        "}}" +
                    "</style>" +
                "</head>" +
                "<body style=\"font-family:sans-serif;\">" +
                    "{0}" + // Put body data here.
                "</body>" +
            "</html>";

        private static string TITLE_HEADER =
            "<h1 style=\"font-size:1.6em;\">" +
                "{0}" + // Put header data here.
            "</h1>";

        private static string MAIN_HEADER =
            "<h2 style=\"font-size:1.2em;\">" +
                "{0}" + // Put header data here.
            "</h2>";

        private static string FILE_HEADER =
            "<h3 style=\"font-size:1.0em;\">" +
                "{0}" + // Put header data here.
            "</h3>";

        private static string ERROR_HEADER =
            "<h4 style=\"font-size:0.8em;\">" +
                "{0}" + // Put header data here.
            "</h4>";

        private static string ERROR_TABLE_BODY =
            "<table border=\"0\" width=\"100%\" cellpadding=\"4\" cellspacing=\"0\">" +
                "{0}" + // Put data rows and columns here.
            "</table>";

        private static string ERROR_TABLE_LINE =
            "<tr valign=\"top\">" +
                "<td class=\"error_table_line\">" +
                    "<i>{0}</i>" + // Put value of left column here.
                "</td>" +
                "<td class=\"error_table_line\">" +
                    "{1}" +  // Put value of error message here.
                    "<br />" +
                    "<span style=\"font-family:monospace;\">" +
                    "{2}" + // Put vlaue of file path here.
                    "</span>" +
                "</td>" +
            "</tr>";

        private static string INFO_TABLE_BODY =
            "<table border=\"0\" width=\"100%\" cellpadding=\"2\" cellspacing=\"0\">" +
                "<colgroup>" +
                    "<col align=\"left\" valign=\"middle\"/>" +
                    "<col align=\"left\" valign=\"middle\"/>" +
                "</colgroup>" +
                "{0}" + // Put data rows and columns here.
            "</table>";

        private static string INFO_TABLE_LINE =
            "<tr>" +
                "<td class=\"info_table_line\" width=\"80\">" +
                    "{0}" + // Put value of left column here.
                "</td>" +
                "<td class=\"info_table_line\">" +
                    "{1}" +  // Put value of right column here.
                "</td>" +
            "</tr>";

        private static string EXPORT_TABLE_BODY =
            "<table border=\"0\" width=\"100%\" cellpadding=\"2\" cellspacing=\"0\">" +
                "<colgroup>" +
                    "<col align=\"left\" valign=\"middle\"/>" +
                    "<col align=\"left\" valign=\"middle\"/>" +
                "</colgroup>" +
                "{0}" + // Put data rows and columns here.
            "</table>";

        private static string EXPORT_TABLE_LINE =
            "<tr>" +
                "<td class=\"export_table_line\" width=\"100\">" +
                    "{0}" + // Put value of left column here.
                "</td>" +
                "<td class=\"export_table_line\">" +
                    "{1}" +  // Put value of right column here.
                "</td>" +
            "</tr>";

        private static string METRIC_TABLE_BODY =
            "<table border=\"0\" width=\"100%\" cellpadding=\"4\" cellspacing=\"0\">" +
                "<colgroup>" +
                    "<col align=\"left\" valign=\"middle\"/>" +
                    "<col align=\"right\" valign=\"middle\"/>" +
                "</colgroup>" +
                "<tr>" +
                    "<th class=\"metric_table_header\">" +
                        "{0}" +  // Put left column header here.
                    "</th>" +
                    "<th class=\"metric_table_header\">" +
                        "{1}" + // Put right column header here. 
                    "</th>" +
                "</tr>" +
                "{2}" + // Put data rows and columns here.
            "</table>";

        private static string METRIC_TABLE_LINE =
            "<tr>" +
                "<td class=\"metric_table_line\">" +
                    "{0}" + // Put value of left column here.
                "</td>" +
                "<td class=\"metric_table_line\">" +
                    "{1}" +  // Put value of right column here.
                "</td>" +
            "</tr>";

        private static string RATIO_TABLE_BODY =
            "<table border=\"0\" width=\"100%\" cellpadding=\"4\" cellspacing=\"0\">" +
                "<colgroup>" +
                    "<col align=\"left\" valign=\"middle\"/>" +
                    "<col align=\"right\" valign=\"middle\"/>" +
                "</colgroup>" +
                "<tr>" +
                    "<th class=\"ratio_table_header\">" +
                        "{0}" +  // Put left column header here.
                    "</th>" +
                    "<th class=\"ratio_table_header\">" +
                        "{1}" + // Put right column header here. 
                    "</th>" +
                "</tr>" +
                "{2}" + // Put data rows and columns here.
            "</table>";

        private static string RATIO_TABLE_LINE =
            "<tr>" +
                "<td class=\"ratio_table_line\">" +
                    "{0}" + // Put value of left column here.
                "</td>" +
                "<td class=\"ratio_table_line\">" +
                    "{1}" +  // Put value of right column here.
                "</td>" +
            "</tr>";

        private static string SCHEME_TABLE_BODY =
            "<table border=\"0\" width=\"100%\" cellpadding=\"4\" cellspacing=\"0\">" +
                "<colgroup>" +
                    "<col align=\"left\" valign=\"middle\"/>" +
                    "<col align=\"right\" valign=\"middle\"/>" +
                    "<col align=\"right\" valign=\"middle\"/>" +
                    "<col align=\"right\" valign=\"middle\"/>" +
                    "<col align=\"right\" valign=\"middle\"/>" +
                "</colgroup>" +
                "<tr>" +
                    "<th class=\"scheme_table_header\" rowspan=\"2\">" +
                        "{0}" +  // Put first column header here.
                    "</th>" +
                    "<th class=\"scheme_table_header\" colspan=\"2\">" +
                        "{1}" +  // Put second column header here.
                    "</th>" +
                    "<th class=\"scheme_table_header\" colspan=\"2\">" +
                        "{4}" +  // Put fifth column header here.
                    "</th>" +
                "</tr>" +
                "<tr>" +
                    "<th class=\"scheme_table_header\" width=\"60\">" +
                        "{2}" +  // Put third column header here.
                    "</th>" +
                    "<th class=\"scheme_table_header\" width=\"60\">" +
                        "{3}" +  // Put fourth column header here.
                    "</th>" +
                    "<th class=\"scheme_table_header\" width=\"60\">" +
                        "{5}" +  // Put sixth column header here.
                   "</th>" +
                    "<th class=\"scheme_table_header\" width=\"60\">" +
                        "{6}" +  // Put seventh column header here.
                    "</th>" +
                "</tr>" +
                "{7}" + // Put data rows and columns here.
            "</table>";

        private static string SCHEME_TABLE_LINE =
            "<tr>" +
                "<td class=\"scheme_table_line\">" +
                    "{0}" + // Put value of first column here.
                "</td>" +
                "<td class=\"scheme_table_line\">" +
                    "{1}" +  // Put value of second column here.
                "</td>" +
                "<td class=\"scheme_table_line\">" +
                    "{2}" + // Put value of third column here.
                "</td>" +
                "<td class=\"scheme_table_line\">" +
                    "{3}" +  // Put value of fourth column here.
                "</td>" +
                "<td class=\"scheme_table_line\">" +
                    "{4}" +  // Put value of fifth column here.
                "</td>" +
            "</tr>";

        private static string FOOTNOTE_ITEM =
            "{0}" +     // Put item label here. 
            "<sup style=\"font-size:0.8em; font-weight:normal;\">" +
                "{1}" + // Put footnote tag here.
            "</sup>";

        private static string FOOTNOTE_TEXT =
            "<span style=\"font-size:0.7em;\">" +
                "<sup>" +
                    "{0}" + // Put footnote tag here.
                "</sup>" +
                "&nbsp;" +
                "{1}" +     // Put footnote text here.
            "</span><br/>";

        private static string LINE_BREAK = "<br/>";

        private static string FIXED_SPACE = "&nbsp;";

        #endregion // HTML tag definition section.
    }
}
