﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Diagnostics;
using System.Collections.Generic;

namespace plexdata.LineCounter
{
    public class ScanningResults : ScanningResult
    {
        public ScanningResults(FileScanner fileScanner)
            : base()
        {
            this.Label = "Summary";
            this.TypeResults = new ScanningResult[] { };
            this.FileScanner = fileScanner;

            try
            {
                if (this.FileScanner != null)
                {
                    List<ScanningResult> patternTypes = new List<ScanningResult>();
                    foreach (string pattern in this.FileScanner.Filter.SearchPattern.Values)
                    {
                        string type = Path.GetExtension(pattern);
                        ContentScanner[] contentScanner = this.FileScanner.ScannersByExtension(type);
                        if (contentScanner.Length > 0)
                        {
                            ScanningResult typeResult = new ScanningResult(("*" + type), contentScanner);
                            this.ScanDuration += typeResult.ScanDuration;
                            this.TotalLines += typeResult.TotalLines;
                            this.EmptyLines += typeResult.EmptyLines;
                            this.CommentLines += typeResult.CommentLines;
                            this.LinesWithCode += typeResult.LinesWithCode;
                            this.StringLiterals += typeResult.StringLiterals;
                            this.CharacterLiterals += typeResult.CharacterLiterals;
                            patternTypes.Add(typeResult);
                        }
                    }
                    this.TypeResults = patternTypes.ToArray();
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        public FileScanner FileScanner { get; private set; }

        public ScanningResult[] TypeResults { get; private set; }
    }

    public class ScanningResult
    {
        protected ScanningResult()
            : base()
        {
            this.Label = String.Empty;
            this.ScanDuration = new TimeSpan();
            this.FilesCount = 0;
            this.TotalLines = 0;
            this.EmptyLines = 0;
            this.CommentLines = 0;
            this.LinesWithCode = 0;
            this.StringLiterals = 0;
            this.CharacterLiterals = 0;
            this.FindingResults = new FindingResult[] { };
        }

        public ScanningResult(string label, ContentScanner[] contentScanner)
            : this()
        {
            this.Label = label;
            try
            {
                if (contentScanner != null && contentScanner.Length > 0)
                {
                    List<FindingResult> findingResults = new List<FindingResult>();

                    this.ContentScanners = new List<ContentScanner>(contentScanner).ToArray();

                    foreach (ContentScanner current in contentScanner)
                    {
                        this.ScanDuration += current.ScanDuration;
                        this.TotalLines += current.TotalLines;
                        this.EmptyLines += current.EmptyLines;
                        this.CommentLines += current.CommentLines;
                        this.LinesWithCode += current.LinesWithCode;
                        this.StringLiterals += current.StringLiterals;
                        this.CharacterLiterals += current.CharacterLiterals;

                        foreach (CommentFindings commentFindings in current.Findings)
                        {
                            string commentScheme = commentFindings.Scheme.ToString();

                            FindingResult findingResult = findingResults.Find(
                                delegate(FindingResult item) { return item.IsEqual(commentFindings.Scheme); });

                            if (findingResult == null)
                            {
                                findingResult = new FindingResult(commentFindings.Scheme);
                                findingResults.Add(findingResult);
                            }

                            findingResult.PureCommentCount += commentFindings.PureCommentCount;
                            findingResult.PureCommentLines += commentFindings.PureCommentLines;
                            findingResult.CodeCommentCount += commentFindings.CodeCommentCount;
                            findingResult.CodeCommentLines += commentFindings.CodeCommentLines;
                        }
                    }
                    this.FilesCount = contentScanner.Length;
                    this.FindingResults = findingResults.ToArray();
                }
            }
            catch (Exception exception)
            {
                Debug.WriteLine(exception);
            }
        }

        public string Label { get; protected set; }

        public TimeSpan ScanDuration { get; protected set; }

        public int FilesCount { get; protected set; }

        public long TotalLines { get; protected set; }

        public long EmptyLines { get; protected set; }

        public long CommentLines { get; protected set; }

        public long LinesWithCode { get; protected set; }

        public long StringLiterals { get; protected set; }

        public long CharacterLiterals { get; protected set; }

        public FindingResult[] FindingResults { get; protected set; }

        public ContentScanner[] ContentScanners { get; protected set; }
    }

    public class FindingResult
    {
        public FindingResult()
            : base()
        {
            this.CommentScheme = String.Empty;
            this.PureCommentCount = 0;
            this.PureCommentLines = 0;
            this.CodeCommentCount = 0;
            this.CodeCommentLines = 0;
        }

        public FindingResult(CommentScheme commentScheme)
            : this()
        {
            this.CommentScheme = FindingResult.BuildCommentScheme(commentScheme);
        }

        public string CommentScheme { get; private set; }

        public int PureCommentCount { get; set; }

        public int PureCommentLines { get; set; }

        public int CodeCommentCount { get; set; }

        public int CodeCommentLines { get; set; }

        public bool IsEqual(CommentScheme commentScheme)
        {
            if (commentScheme != null)
            {
                return String.Compare(
                    this.CommentScheme,
                    FindingResult.BuildCommentScheme(commentScheme)) == 0;
            }
            else
            {
                return false;
            }
        }

        public static string BuildCommentScheme(CommentScheme commentScheme)
        {
            string result = String.Empty;

            if (commentScheme != null && !String.IsNullOrEmpty(commentScheme.Open))
            {
                result = commentScheme.Open;
                if (!String.IsNullOrEmpty(commentScheme.Close))
                {
                    result += " ... " + commentScheme.Close;
                }
            }
            return result.Trim();
        }
    }
}
