﻿/*
 * Copyright (C)  2012  Axel Kesseler
 * 
 * This software is free and you can use it for any purpose. Furthermore, 
 * you are free to copy, to modify and/or to redistribute this software.
 * 
 * In addition, this software is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

using System;
using System.IO;
using System.Diagnostics;
using System.Collections.Generic;

namespace plexdata.LineCounter
{
    public class SearchPattern : Object, ICloneable, IComparable, IComparable<string>, IComparable<SearchPattern>, IComparer<string>, IComparer<SearchPattern>
    {
        private static string FORMAT = "*.{0}";
        private static string DELIMITER = "; ";
        private static char[] SPLITS = new char[] { ' ', ';', ',', ':', '.', '*', '?', '"', '<', '>', '|', '\t', '\r', '\n', '\\', '/' };

        public SearchPattern()
            : base()
        {
            this.Values = new string[0];
        }

        public SearchPattern(string items)
            : this()
        {
            this.Parse(items);
        }

        public SearchPattern(string[] items)
            : this()
        {
            this.Parse(items);
        }

        private string[] values = null;
        public string[] Values
        {
            get { return this.values; }
            set { this.Parse(value); }
        }

        public bool Parse(string items)
        {
            bool success = false;
            List<string> helper = new List<string>();
            if (items != null)
            {
                // Ok, let's summarize. After a couple of tests it came up that splitting 
                // input strings like *.* (and all combinations) using this split operation 
                // the result string are empty. This might be a problem... But on the other 
                // hand using all files for an analysis is really not a good idea. 
                //
                // Therefore, grepping in all files is not allowed by definition!
                //
                foreach (string current in items.Split(SPLITS, StringSplitOptions.RemoveEmptyEntries))
                {
                    // Use lower cases and suppress duplicates...
                    string candidate = String.Format(FORMAT, current).ToLower();
                    if (!helper.Contains(candidate))
                    {
                        helper.Add(candidate);
                    }
                }

                // Don't use the property!
                this.values = helper.ToArray();
                success = this.values.Length > 0;
            }
            return success;
        }

        public bool Parse(string[] items)
        {
            string helper = String.Empty;
            if (items != null)
            {
                foreach (string curren in items)
                {
                    helper += curren + SPLITS[0];
                }
            }
            return this.Parse(helper);
        }

        public void Sort()
        {
            this.Sort(true);
        }

        public void Sort(bool ascending)
        {
            if (this.values != null && this.values.Length > 0)
            {
                Array.Sort<string>(this.values);
                if (!ascending)
                {
                    Array.Reverse(this.values);
                }
            }
        }

        public object Clone()
        {
            SearchPattern clone = new SearchPattern();
            if (this.Values != null)
            {
                List<string> helper = new List<string>();
                foreach (string value in this.Values)
                {
                    string current = value.Clone() as string;
                    if (current != null)
                    {
                        helper.Add(current);
                    }
                }
                clone.values = helper.ToArray();
            }
            return clone;
        }

        public int CompareTo(object other)
        {
            if (other is string)
            {
                return this.CompareTo(other as string);
            }
            else if (other is SearchPattern)
            {
                return this.CompareTo(other as SearchPattern);
            }
            else
            {
                throw new ArgumentException();
            }
        }

        public int CompareTo(string other)
        {
            if (other != null)
            {
                return String.Compare(this.ToString(), other);
            }
            else
            {
                throw new ArgumentNullException();
            }
        }

        public int CompareTo(SearchPattern other)
        {
            if (other != null)
            {
                return String.Compare(this.ToString(), other.ToString());
            }
            else
            {
                throw new ArgumentNullException();
            }
        }

        public int Compare(string left, string right)
        {
            return String.Compare(left, right);
        }

        public int Compare(SearchPattern left, SearchPattern right)
        {
            if (left != null && right != null)
            {
                return this.Compare(left.ToString(), right.ToString());
            }
            else if (left == null && right != null)
            {
                return -1;
            }
            else if (left != null && right == null)
            {
                return 1;
            }
            else
            {
                throw new ArgumentNullException();
            }
        }

        public override string ToString()
        {
            string result = String.Empty;
            foreach (string pattern in this.Values)
            {
                result += pattern + DELIMITER;
            }

            // Cutoff last delimiter...
            if (result.Length >= DELIMITER.Length)
            {
                result = result.Substring(0, result.Length - DELIMITER.Length);
            }

            return result;
        }
    }
}
